"""_summary_
"""
from __future__ import annotations
from typing import List
from .verifyta import Verifyta
from .umodel import UModel
from .iTools import Mermaid


def set_verifyta_path(verifyta_path: str):
    """Set verifyta path, and you will get tips if `verifyta_path` is invalid.
    This function will check whether `verifyta_path` is valid by following steps:
    1. run '{verifyta_path} -h' with cmd
    2. check whether '-h [ --help ]' is in the result

    Args:
        verifyta_path (str): absolute path to `verifyta`
    """
    Verifyta().set_verifyta_path(verifyta_path)


def easy_verify(model_path: str | List[str],
                trace_path: str | List[str],
                verify_options: str = "-t 1",
                num_threads=1) -> List[str]:
    """Easy verification with default options, return to the shortest diagnostic path.
    Verify the model in model_path and save the verification results to trace_path.

    Args:
        model_path (str | List[str]): model paths to be verified.
        trace_path (str | List[str], optional): target trace paths, both `.xtr` and `.xml`(DBM) are supported.
            Defaults to None, which will create `.xtr` path.
        verify_options (str | List[str], optional): verify options that are proveded by `verifyta`, and you can get details by run `verifyta -h` in your terminal.
            If `verify_options` is provides as a single string, all the models will be verified with the same options. Defaults to '-t 1', returning the shortest trace.
        num_threads (int, optional): use multi-threads if is greater than 1. Defaults to 1.

    Returns:
        List[str]: terminal verify results for each `.xml` model.
    """
    return Verifyta().easy_verify(model_path, trace_path, verify_options, num_threads)

def get_communication_graph(model_path: str, save_path=None, is_beautify=True) -> Mermaid:
    """Get the communication graph of the UPPAAL model, and return a `Mermaid` instance.

    Args:
        model_path (str): path to the model.
        save_path (_type_, optional): `<.md | .svg | .pdf | .png>`, the path to save the file. Defaults to None.
        is_beautify (bool, optional): whether beautify the mermaid file by merging edges. Defaults to True.

    Returns:
        Mermaid: _description_
    """
    umod = UModel(model_path)
    return umod.get_communication_graph(save_path, is_beautify)


def verify(model_path: str | List[str],
           verify_options: str | List[str] = None,
           num_threads: int = 1) -> List[str]:
    """Verify models and return the verify results as list. This is designed for advanced UPPAAL user. 
    If you want to save a `.xtr` or `.xml`(DBM) path, you may want to check `Verifyta().easy_verify()`.
    WARNING: Note that `-f xx.xtr` or `-X xx.xml` should be used together with `-t` options, otherwise you may fail to save the path files.

    Examples:
        >>> Verifyta().verify('test1.xml')
        >>> Verifyta().verify(['test1.xml', 'test2.xml'], verify_options = '-t 1 -o 0')
        >>> Verifyta().verify(['test1.xml', 'test1.xml'], 
        >>>     verify_options = ['-t 1 -o 0', '-t 2 -o 0'])
        >>> # if you surely want to generate a trace file with Verifyta().verify()
        >>> # you should not add `.xtr` at the end of `xtr_trace`,
        >>> # or `.xml` at the end of `xtr_trace`
        >>> Verifyta().verify(['test1.xml', 'test1.xml'],
        >>>     verify_options = ['-f xtr_trace -t 1 -o 0', '-X xml_trace -t 2 -o 0'],
        >>>     num_threads=2)
        >>>
        >>> # return example
        >>> # Options for the verification:
        >>> #    Generating shortest trace
        >>> #    Search order is breadth first
        >>> #    Using conservative space optimisation
        >>> #    Seed is 1665658616
        >>> #    State space representation uses minimal constraint systems
        >>> #    Verifying formula 1 at /nta/queries/query[1]/formula
        >>> #   -- Formula is satisfied.
        >>> # Options for the verification:
        >>> #   Generating shortest trace
        >>> #   Search order is breadth first
        >>> #   Using conservative space optimisation
        >>> #   Seed is 1665658616
        >>> #   State space representation uses minimal constraint systems
        >>> #   Verifying formula 1 at /nta/queries/query[1]/formula
        >>> #   -- Formula is NOT satisfied.


    Args:
        model_path (str | List[str]): model paths to be verified.
        verify_options (str | List[str], optional): verify options that are proveded by `verifyta`, and you can get details by run `verifyta -h` in your terminal.
            If `verify_options` is provided as a single `string`, all the models will be verified with the same options. Defaults to None.
        num_threads (int, optional): use multi-threads if is greater than 1. Defaults to 1.

    Raises:
        ValueError: _description_
        TypeError: _description_
        ValueError: _description_
        FileNotFoundError: _description_
        ValueError: _description_

    Returns:
        List[str]: terminal verify results for each `.xml` model. 
    """
    return Verifyta().verify(model_path, verify_options, num_threads)


def cmd(cmd: str) -> str:
    """Run common command with cmd, you can easily ignore the verifyta path.

    Args:
        cmd (str): command to be run by verifyta.

    Returns:
        str: the running cmd and the command result
    """
    return Verifyta().cmd(cmd)


def cmds(cmds: List[str], num_threads: int = 1) -> List[str]:
    """Run commands with terminal.

    Args:
        cmds (List[str]): commands to run
        num_threads (int, optional): use multi-threads if is greater than 1. Defaults to 1.

    Raises:
        ValueError: Number of threads should ≥ 1.

    Returns:
        List[str]: return values of each command.
    """
    return Verifyta().cmds(cmds, num_threads)



"""命令行输入pytest -s可以查看print
"""
import os
import time
import pytest
import verifyta_path
from pyuppaal import Verifyta

Verifyta().set_verifyta_path(verifyta_path.VERIFYTA_PATH)


def test_set_verifyta_path():
    """_summary_
    """
    Verifyta().set_verifyta_path(verifyta_path.VERIFYTA_PATH)
    with pytest.raises(ValueError) as excinfo:
        Verifyta().set_verifyta_path('invalid path')
    assert Verifyta().set_verifyta_path(verifyta_path.VERIFYTA_PATH) is True


def test_verify():
    """_summary_
    """
    Verifyta().set_verifyta_path(verifyta_path.VERIFYTA_PATH)
    model_paths = [verifyta_path.bring_to_root('demo1.xml'),
                   verifyta_path.bring_to_root('demo2.xml'),
                   verifyta_path.bring_to_root('demo3.xml')]
    # single thread loop
    t0 = time.time()
    verify_ress = Verifyta().verify(model_paths, verify_options='-t 1 -o 0')
    for verify_res in verify_ress:
        assert 'satisfied' in verify_res
    t1 = time.time()
    # multi-thread
    verify_ress = Verifyta().verify(model_paths, verify_options=[
        '-t 1 -o 0', '-t 2 -o 0', '-t 2 -o 1'], num_threads=3)
    for verify_res in verify_ress:
        assert 'satisfied' in verify_res
    t2 = time.time()
    # time usage
    single_thread_time = round(t1-t0, 2)
    multi_thread_time = round(t2-t1, 2)
    assert single_thread_time > multi_thread_time
    info = f'3个复杂度相似单不完全相同任务, 单线程: {single_thread_time}s,'
    info += f'单线程平均: {round(single_thread_time/3, 2)}s, 3线程: {multi_thread_time}s.'
    print(info)


def test_easy_verify1():
    """_summary_
    """
    Verifyta().set_verifyta_path(verifyta_path.VERIFYTA_PATH)
    model_paths = [verifyta_path.bring_to_root('demo1.xml'),
                   verifyta_path.bring_to_root('demo2.xml'),
                   verifyta_path.bring_to_root('demo3.xml')]
    # ======== without trace_paths ========
    # the third model do not have counter-example
    # should automatically create .xtr counter-example
    res = Verifyta().easy_verify(model_paths, verify_options=[
        '-t 1 -o 0', '-t 2 -o 0', '-t 2 -o 1'], num_threads=3)
    target_trace_paths = [verifyta_path.bring_to_root('demo1-1.xtr'),
                          verifyta_path.bring_to_root('demo2-1.xtr')]
    for trace_path in target_trace_paths:
        assert os.path.exists(trace_path) is True
        os.remove(trace_path)


def test_easy_verify2():
    """_summary_
    """
    Verifyta().set_verifyta_path(verifyta_path.VERIFYTA_PATH)
    model_paths = [verifyta_path.bring_to_root('demo1.xml'),
                   verifyta_path.bring_to_root('demo2.xml'),
                   verifyta_path.bring_to_root('demo3.xml')]
    # ======== with trace_paths ========
    trace_paths = [verifyta_path.bring_to_root('t1.xtr'),
                   verifyta_path.bring_to_root('t2-.xml'),
                   verifyta_path.bring_to_root('t3-.xml')]
    res = Verifyta().easy_verify(model_paths, trace_path=trace_paths,
                                 verify_options='-t 1 -o 0', num_threads=3)
    print(res)
    # the third model do not have counter-example
    target_trace_paths = [verifyta_path.bring_to_root('t1-1.xtr'),
                          verifyta_path.bring_to_root('t2-1.xml')]
    for trace_path in target_trace_paths:
        assert os.path.exists(trace_path) is True
        os.remove(trace_path)



from typing import List, Tuple, Dict
import xml.etree.cElementTree as ET


class UFactory:
    """
    Generate xml elements of UPPAAL, including queries, declaration, location, transition, template and so on.
    """
    @staticmethod
    def __query(query: str) -> ET.Element:
        """Components of `UFactory.queries()`.

        >>> <query>
        >>>     <formula>E&lt;&gt; Output.pass</formula>
        >>>     <comment />
        >>> </query>

        Args:
            query (str): _description_

        Returns:
            ET.Element: _description_
        """
        query_elem = ET.Element('query')

        # 构建并添加formula
        formula_elem = ET.Element('formula')
        formula_elem.text = query
        query_elem.append(formula_elem)

        # 构建并添加comment
        query_elem.append(ET.Element('comment'))
        return query_elem

    @staticmethod
    def queries(queries: List[str]) -> ET.Element:
        """_summary_

        >>> <queries>
        >>>     <query>
        >>>         <formula>E&lt;&gt; Output.pass</formula>
        >>>         <comment />
        >>>     </query>
        >>>     <query>
        >>>         <formula>E&lt;&gt; Output2.pass</formula>
        >>>         <comment />
        >>>     </query>
        >>> </queries>

        Args:
            queries (List[str]): _description_

        Returns:
            ET.Element: _description_
        """
        queries_elem = ET.Element('queries')
        # 构建并加入多个queries element
        for query in queries:
            queries_elem.append(UFactory.__query(query))
        return queries_elem

    # @staticmethod
    # def declaration(declaration: str) -> ET.Element:
    #     """Generate the declararion of UPPAAL.

    #     Args:
    #         declaration (str): all contents of declaration.

    #     Returns:
    #         ET.Element: _description_
    #     """
    #     dec_elem = ET.Element('declaration')
    #     dec_elem.text = declaration
    #     return dec_elem

    @staticmethod
    def location(location_id: int, pos_x: int, pos_y: int, inv: str = None, 
                 name: str = None, is_committed: bool = False) -> ET.Element:
        """
        Generate a location

        :param int id: the `id` of location
        :param int pos_x,pos_y: the position of location
        :param str inv: theinvariant of location, e.g., `gclk<=10` and `gclk<100`.
        :param str name: the name of location
        :param bool is_committed: determine whether it is committed or not


        >>> <location id="id37" x="-169" y="-59">
        >>>     <name x="90" y="166">pass</name>
        >>>     <label kind="invariant" x="-212" y="-42">gclk&lt;=122</label>
        >>>     <committed/>
        >>> </location>
        """
        location = ET.Element('location', {'id': f'id{location_id}',
                                           'x': str(pos_x),
                                           'y': str(pos_y)})
        # 添加名字
        if name is not None:
            location_name = ET.Element('name', {'x': str(pos_x),
                                                'y': str(pos_y - 20)})
            location_name.text = name
            location.append(location_name)
        # 添加inv
        if inv is not None:
            label = ET.Element('label', {'kind': 'invariant',
                                         'x': str(pos_x),
                                         'y': str(pos_y - 40)})
            label.text = inv
            location.append(label)
        # 添加committed
        if is_committed:
            location.append(ET.Element('committed'))
        return location

    @staticmethod
    def transition(sourceID: int, targetID: int, pos_x: int, pos_y: int, 
                   guard: str = None, sync: str = None, clock_reset: str = None, 
                   nail: bool = False) -> ET.Element:
        """
        :param int sourceID: the id of start
        :param int targetID: the id of end
        :param int post_x,post_y: the position of sign
        :param str guard: the guard of the edge, e.g., `t>=10`, `t>100`
        :param str sync: the signal of synchronisation, e.g., `!` and `?`
        :param str clock_reset: e.g., `t=0`
        :param bool nail: determine whether the edge is curved

        >>> <transition>
        >>>     <source ref="id7"/>
        >>>     <target ref="id6"/>
        >>>     <label kind="guard" x="139" y="90">t&gt;=tERPMin</label>
        >>>     <label kind="synchronisation" x="123" y="90">sync?</label>
        >>>     <label kind="assignment" x="156" y="107">t=0</label>
        >>> </transition>
        """
        transition = ET.Element('transition')
        # 构建并添加source
        source = ET.Element('source', {'ref': f'id{sourceID}'})
        transition.append(source)
        # 构建并添加target
        target = ET.Element('target', {'ref': f'id{targetID}'})
        transition.append(target)
        # 构建并添加guard
        if guard is not None:
            label_guard = ET.Element('label', {'kind': 'guard',
                                               'x': str(pos_x),
                                               'y': str(pos_y)})
            label_guard.text = guard
            transition.append(label_guard)
        # 构建并添加synchronisation
        if sync is not None:
            label_sync = ET.Element('label', {'kind': 'synchronisation',
                                              'x': str(pos_x),
                                              'y': str(pos_y - 30)})
            label_sync.text = sync
            transition.append(label_sync)
        # 构建并添加assignment
        if clock_reset is not None:
            label_assignment = ET.Element('label', {'kind': 'assignment',
                                                    'x': str(pos_x),
                                                    'y': str(pos_y - 60)})
            label_assignment.text = clock_reset
            transition.append(label_assignment)
        # 构建并添加nail (弯曲结点)
        if nail:
            nail_element = ET.Element('nail', {'x': str(pos_x),
                                               'y': str(pos_y)})
            transition.append(nail_element)
        return transition

    @staticmethod
    def template(name: str, locations: List[ET.Element], init_id: int, transitions: List[ET.Element],
                 parameter: str = None, declaration: str = None) -> ET.Element:
        """
        :param str name: the name of template
        :param List[ET.Element] locations: series of locations generated by UFactory.construct_location()
        :param int init_id: the id of initial location
        :param List[ET.Element] transitions: series of transitions generated by UFactory.construct_location()
        :param str parameter: parameter instruction, e.g., <parameter>broadcast chan &amp;actNode, actPath,int tERPMin, int tERPMax</parameter>
        :param str declaration: declaration instruction, e.g., <declaration>clock t;</declaration>

        >>> <template>
        >>>     <name>NodeN</name>
        >>>     <parameter>broadcast chan &amp;actNode, int tERPMin, int tERPMax</parameter>
        >>>     <declaration>clock t;</declaration>
        >>>     <location> </location>
        >>>     <location> </location>
        >>>     <location> </location>
        >>>     <init ref="id47"/>
        >>>     <transition> </transition>
        >>>     <transition> </transition>
        >>>     <transition></transition>
        >>> </template>
        """
        template = ET.Element('template')
        # 创建并加入name
        name_elem = ET.Element('name')
        name_elem.text = name
        template.append(name_elem)

        # 创建并加入parameter
        if parameter is not None:
            parameter_elem = ET.Element('parameter')
            parameter_elem.text = parameter
            template.append(parameter_elem)

        # 创建并加入declaration
        if declaration is not None:
            declaration_elem = ET.Element('declaration')
            declaration_elem.text = declaration
            template.append(declaration_elem)

        # 加入locations
        for location in locations:
            template.append(location)

        # 创建并加入init
        init_elem = ET.Element('init', {'ref': f'id{init_id}'})
        template.append(init_elem)

        # 加入 transitions
        for transition in transitions:
            template.append(transition)
        return template

    @staticmethod
    def input(input_name: str, signals: List[Tuple[str, str, str]], init_id: int) -> ET.Element:
        """
        Generate a linear model of input.

        :param str input_name: the name of input
        :parama List[Tuple[str, int, int]] signals: this data type is updated in `umodel.TymedActions`
        :param int init_id: int, set the minimum `<init ref='xxx'/>` of monitor
        """
        # 创建locations
        locations = []
        for i in range(len(signals)):
            # [signal, guard, inv, name]
            location = UFactory.location(init_id + i, 300 * i, 200, signals[i][2])
            locations.append(location)
        # 需要多一个尾巴location
        location = UFactory.location(init_id + len(signals), 300 * len(signals), 200, name='pass')
        locations.append(location)
        # 创建transitions
        transitions = []
        for i in range(len(signals)):
            # [signal, guard, inv]
            transition = UFactory.transition(init_id + i, init_id + i + 1, i * 300 + 100, 200,
                                                  signals[i][1], signals[i][0]+'!')
            transitions.append(transition)
        # 获得clock name并创建declaration
        clk_name = signals[0][1].split('>')[0]
        declaration = f'clock {clk_name};'
        monitor_elem = UFactory.template(input_name, locations, init_id, transitions,declaration=declaration)
        return monitor_elem

    @staticmethod
    def monitor(monitor_name: str, signals: List[Tuple[str, str, str]], observe_action: List[str], 
                init_id: int, strict: bool = False, allpattern: bool=False) -> ET.Element:
        """
        Generate a new linear model of.

        :param str monitor_name: the name of monitor
        :param List[Tuple[str, int, int]] signals: this data type is updated in `umodel.TimedActions`
        :param int init_id: set the minimum `<init ref='xxx'/>` of monitor
        :param bool strict: determine whethe the observation is strict, which indicates other observations will be forbidden

        >>> <template>
        >>>     <name>Monitor</name>
        >>>     <location> </location>
        >>>     <location>init ref="id47"/>
        >>>     <transition> </transition>
        >>> </template>
        """
        # 创建locations
        locations = []
        for i in range(len(signals)):
            # [signal, guard, inv]
            inv = None if allpattern else signals[i][2]
            location = UFactory.location(init_id + i, 300 * i, 200, inv=inv)
            locations.append(location)
        # 需要多一个尾巴location
        location = UFactory.location(init_id + len(signals), 300 * len(signals), 200, name='pass')
        locations.append(location)

        # 创建transitions
        transitions = []
        for i in range(len(signals)):
            # [signal, guard, inv]
            guard = None if allpattern else signals[i][1]
            transition = UFactory.transition(init_id + i, init_id + i + 1, i * 300 + 100, 200,
                                                  guard=guard, sync=signals[i][0]+'?')
            transitions.append(transition)

        # 如果是strict，需要给每个location创建fail location，并且构建对应transitions边指向对应的fail location
        if strict:
            pruned_observation_action = list(set(observe_action))

            # 构建指向fail的transitions
            for i in range(len(signals)):
                # 构建一个fail location
                fail_location_id = init_id + len(signals) + i + 1
                # check allpattern
                inv = None if allpattern else signals[i][2].replace('<=', '<')
                location = UFactory.location(fail_location_id, 300 * i, -200, inv=inv,
                                                  name='fail' + str(i))
                locations.append(location)
                actions_length = 200 // len(pruned_observation_action)
                # 需要对每一个observe action指向fail transitions
                for j in range(len(pruned_observation_action)):
                    # 指向fail，注意guard是大于
                    guard = None if allpattern else signals[i - 1][1].replace('>=', '>')
                    if allpattern and (pruned_observation_action[j] == signals[i][0]):
                        continue
                    if i == 0:
                        transition = UFactory.transition(init_id + i, fail_location_id, i * 300 + (j-len(pruned_observation_action)//2) * actions_length, -100,
                                                         guard=None, sync=pruned_observation_action[j]+'?',nail=True)
                    else:
                        transition = UFactory.transition(init_id + i, fail_location_id, i * 300 + (j-len(pruned_observation_action)//2) * actions_length, -100,
                                                         guard=guard, sync=pruned_observation_action[j]+'?',nail=True)

                    transitions.append(transition)
        # 获得clock name并创建declaration
        clk_name = signals[0][1].split('>')[0]
        declaration = None if allpattern else f'clock {clk_name};'
        monitor_elem = UFactory.template(monitor_name, locations, init_id, transitions, declaration=declaration)
        return monitor_elem



"""verifyta
"""
# support typing str | List[str]
# https://github.com/microsoft/pylance-release/issues/513
from __future__ import annotations

from typing import List
from multiprocessing.dummy import Pool as ThreadPool
import functools
import platform
import os


def check_is_verifyta_path_empty(func):
    """====该函数后续会删除，希望不要每次都检测====
    装饰器, 用来在verifyta运行前检测路径是否被设置。
    """
    @functools.wraps(func)
    def checker_wrapper(*args, **kwargs):
        # 注意verifyta是singleton，因此可以直接用Verifyta()调用到唯一的实例
        if Verifyta().verifyta_path:
            return func(*args, **kwargs)
        else:
            error_info = 'Verifyta path is not set.'
            error_info += ' Please use "pyuppaal.set_verifyta_path(verifyta_path: str)" to set the path of verifyta.'
            raise ValueError(error_info)
    return checker_wrapper


class Verifyta:
    """This is a singleton class that help to use `verifyta` command.
    """
    # make singleton
    __instance = None
    __is_first_init = True

    def __new__(cls):
        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
        return cls.__instance

    def __init__(self):
        # make singleton
        if not self.__is_first_init:
            return
        self.__is_first_init = False

        self.__verifyta_path: str = None
        self.__is_windows: bool = platform.system() == 'Windows'

    @property
    def verifyta_path(self) -> str:
        """Get current verifyta path.

        Returns:
            str: current verifyta path.
        """
        return self.__verifyta_path

    def __is_valid_verifyta_path(self, verifyta_path: str) -> bool:
        """Check whether the verifyta path is valid.
        Checking steps:
        1. run cmd with `verifyta_path -h`;
        2. check whether `'-h [ --help ]' in res`.

        Args:
            verifyta_path (str): _description_

        Returns:
            bool: _description_
        """
        cmd = f'{verifyta_path} -h'
        res = os.popen(cmd).read()
        return '-h [ --help ]' in res

    def set_verifyta_path(self, verifyta_path: str) -> bool:
        """Set the verifyta path before using pyuppaal.
        This function will check the validation of the `verifyta_path` automatically by the following steps:

        1. run cmd with `verifyta_path -h`;
        2. check whether `'-h [ --help ]' in res`.
        
        Example paths:

        1. Windows: path_to_uppaal\\bin-Windows\\verifyta.exe
        2. Linux  : path_to_uppaal/bin-Linux/verifyta
        3. macOS  : path_to_uppaal/bin-Darwin/verifyta

        Args:
            verifyta_path (str): (absolute) path to `verifyta`

        Raises:
            ValueError: if verifyta_path is invalid, it will raise an error and tips to help.

        Returns:
            bool: return `True` if verifyta_path is successfully set.
        """

        if self.__is_valid_verifyta_path(verifyta_path):
            self.__verifyta_path = verifyta_path
            return True
        else:
            example_info = "======== Example Paths ========" \
                           "\nWindows: absolute_path_to_uppaal\\bin-Windows\\verifyta.exe" \
                           "\nLinux  : absolute_path_to_uppaal/bin-Linux/verifyta" \
                           "\nmacOS  : absolute_path_to_uppaal/bin-Darwin/verifyta"
            raise ValueError(
                f"Invalid verifyta_path: {verifyta_path}.\n{example_info}")

    @check_is_verifyta_path_empty
    def cmd(self, cmd: str) -> str:
        """Run common command with cmd, you can easily ignore the verifyta path.

        Args:
            cmd (str): command to run.

        Returns:
            str: the output of the input command.
        """
        if self.__verifyta_path not in cmd:
            cmd = f'{self.__verifyta_path} {cmd}'
        return os.popen(cmd).read()

    # @check_is_verifyta_path_empty 调用了self.cmd，所以不需要加
    def cmds(self, cmds: List[str], num_threads: int = 1) -> List[str]:
        """Run commands with terminal.

        Args:
            cmds (List[str]): commands to run
            num_threads (int, optional): use multi-threads if is greater than 1. Defaults to 1.

        Raises:
            ValueError: Number of threads should ≥ 1.

        Returns:
            List[str]: return values of each command.
        """
        # 检测报错
        num_threads = int(num_threads)
        if num_threads < 1:
            raise ValueError("Number of threads should ≥ 1.")
        elif num_threads == 1:
            return [self.cmd(tmp_cmd) for tmp_cmd in cmds]
        else:
            p = ThreadPool(num_threads)
            return p.map(self.cmd, cmds)

    @check_is_verifyta_path_empty
    def _compile_to_if(self, model_path: str) -> str:
        """Compile the `.xml` model_path to a `.if` file and return the path to `.if` file.

        Args:
            model_path (str): `.xml` model file.

        Raises:
            FileNotFoundError: `model_path` not found.
            ValueError: `model_path` is not a `.xml`.
            FileNotFoundError: _description_

        Returns:
            str: path to `.if` file.
        """
        if not os.path.exists(model_path):
            error_info = f'model_path {model_path} not found.'
            raise FileNotFoundError(error_info)
        file_path, file_ext = os.path.splitext(model_path)
        if_path = file_path + '.if'

        if file_ext != '.xml':
            error_info = f'model_path {model_path} should be xml format file.'
            raise ValueError(error_info)

        # set uppaal environment variables
        # 设置命令行环境保证uppaal能够产生正确的.if文件，后半部分保证文件以UTF-8编码，进而保证lf结尾。
        cmd_env = "set UPPAAL_COMPILE_ONLY=1 && set PSDefaultParameterValues['Out-File:Encoding']='Default'"
        # cmd = cmd_env+"&&"+f'{self.__verifyta_path} {model_path} > {if_path}'
        cmd = f'{cmd_env} && {self.__verifyta_path} {model_path} > {if_path}'
        self.cmd(cmd=cmd)
        if not os.path.exists(if_path):
            error_info = f'if file {if_path} has not generated.'
            raise FileNotFoundError(error_info)
        return if_path

    @check_is_verifyta_path_empty
    def compile_to_if(self, model_path: str) -> str:
        """Compile the `.xml` model_path to a `.if` file and return the content of the `.if` file.

        Args:
            model_path (str): `.xml` model file.

        Raises:
            FileNotFoundError: `model_path` not found.
            ValueError: `model_path` is not a `.xml`.
            FileNotFoundError: _description_

        Returns:
            str: path to `.if` file.
        """
        if not os.path.exists(model_path):
            error_info = f'model_path {model_path} not found.'
            raise FileNotFoundError(error_info)
        file_path, file_ext = os.path.splitext(model_path)

        if file_ext != '.xml':
            error_info = f'model_path {model_path} should be xml format file.'
            raise ValueError(error_info)

        # set uppaal environment variables
        # 设置命令行环境保证uppaal能够产生正确的.if文件，后半部分保证文件以UTF-8编码，进而保证lf结尾。
        cmd_env = "set UPPAAL_COMPILE_ONLY=1 && set PSDefaultParameterValues['Out-File:Encoding']='Default'"
        cmd = f'{cmd_env} && {self.__verifyta_path} {model_path}'
        if_str = self.cmd(cmd=cmd)
        return if_str

    # @check_is_verifyta_path_empty 调用了self.cmd，所以不需要加
    def verify(self,
               model_path: str | List[str],
               verify_options: str | List[str] = None,
               num_threads: int = 1) -> List[str]:
        """Verify models and return the verify results as list.
        This is designed for advanced UPPAAL user.
        If you want to save a `.xtr` or `.xml`(DBM) path, you may want to check `Verifyta().easy_verify()`.
        WARNING: Note that `-f xx.xtr` or `-X xx.xml` should be used together with `-t` options, otherwise you may fail to save the path files.
        
        Examples:       
            >>> Verifyta().verify('test1.xml')
            >>> Verifyta().verify(['test1.xml', 'test2.xml'], verify_options = '-t 1 -o 0')
            >>> Verifyta().verify(['test1.xml', 'test1.xml'], 
            >>>     verify_options = ['-t 1 -o 0', '-t 2 -o 0'])
            >>> # if you surely want to generate a trace file with Verifyta().verify()
            >>> # you should not add `.xtr` at the end of `xtr_trace`,
            >>> # or `.xml` at the end of `xtr_trace`
            >>> Verifyta().verify(['test1.xml', 'test1.xml'], 
            >>>     verify_options = ['-f xtr_trace -t 1 -o 0', '-X xml_trace -t 2 -o 0'], 
            >>>     num_threads=2)
            >>>
            >>> # return example
            >>> # Options for the verification: 
            >>> #    Generating shortest trace
            >>> #    Search order is breadth first
            >>> #    Using conservative space optimisation
            >>> #    Seed is 1665658616
            >>> #    State space representation uses minimal constraint systems
            >>> #    Verifying formula 1 at /nta/queries/query[1]/formula
            >>> #   -- Formula is satisfied.
            >>> # Options for the verification: 
            >>> #   Generating shortest trace  
            >>> #   Search order is breadth first
            >>> #   Using conservative space optimisation
            >>> #   Seed is 1665658616  
            >>> #   State space representation uses minimal constraint systems
            >>> #   Verifying formula 1 at /nta/queries/query[1]/formula
            >>> #   -- Formula is NOT satisfied.

        Args:
            model_path (str | List[str]): model paths to be verified.
            verify_options (str | List[str], optional): verify options that are proveded by `verifyta`, and you can get details by run `verifyta -h` in your terminal.
                If `verify_options` is provided as a single `string`, all the models will be verified with the same options. Defaults to None.
            num_threads (int, optional): use multi-threads if is greater than 1. Defaults to 1.

        Raises:
            ValueError: _description_
            TypeError: _description_
            ValueError: _description_
            FileNotFoundError: _description_
            ValueError: _description_

        Returns:
            List[str]: terminal verify results for each `.xml` model. 
        """
        num_threads = int(num_threads)
        if num_threads < 1:
            raise ValueError(
                f"Number of threads should ≥ 1. Current value: {num_threads}.")

        # check model_path is only one model
        if isinstance(model_path, str):
            model_path = [model_path]
        len_model_path = len(model_path)

        # check verify_options is only one
        if verify_options is None:
            verify_options = ['' for _ in range(len_model_path)]
        elif isinstance(verify_options, str):
            verify_options = [verify_options for _ in range(len_model_path)]
        else:
            pass
        len_verify_options = len(verify_options)

        # check consistency
        if len_model_path != len_verify_options:
            error_info = f'Length of model_path and verify_options are inconsistent. Current values:\n' \
                         f'model_path: {model_path}, \nverify_options: {verify_options}.'
            raise ValueError(error_info)

        # set uppaal environment variables
        # 因为生成.if的时候UPPAAL_COMPILE_ONLY=1, 这里要改回来。但是改成啥都不对，所以就啥都不加，然后就对了。。。
        # 啥都不加是@yhc试出来的
        cmd_env = 'set UPPAAL_COMPILE_ONLY=' if self.__is_windows else "UPPAAL_COMPILE_ONLY="

        cmds = []
        for i in range(len_model_path):
            model_i = model_path[i]
            verify_options_i = verify_options[i]
            # check model_path exist
            if not os.path.exists(model_i):
                error_info = f'model_path {model_i} not found.'
                raise FileNotFoundError(error_info)
            # 构造命令
            cmd = cmd_env+'&&' + \
                f'{self.__verifyta_path} {model_i} {verify_options_i}'
            cmds.append(cmd)
        # print(cmds)
        res = self.cmds(cmds=cmds, num_threads=num_threads)
        # print(res)
        return res

    # @check_is_verifyta_path_empty 调用了self.cmd，所以不需要加
    def easy_verify(self,
                    model_path: str | List[str],
                    trace_path: str | List[str] = None,
                    verify_options: str | List[str] = "-t 1",
                    num_threads: int = 1) -> List[str]:
        """Verify models and return the verify results as list.
        For `trace_path` param, both `.xtr` and `.xml`(DBM) files are supported.
        WARNING: `-t` option must be set for `verify_options`, which is set by default `verify_options = '-t 1'`(shortest), otherwise counter-example file may not be created.

        Examples:       
            >>> Verifyta().set_verifyta_path(VERIFYTA_PATH)
            >>> model_paths = [os.path.join(ROOT_DIR, 'demo1.xml'),
            >>>         os.path.join(ROOT_DIR, 'demo2.xml'),
            >>>         os.path.join(ROOT_DIR, 'demo3.xml')]
            >>> trace_paths = [os.path.join(ROOT_DIR, 't1.xtr'),
            >>>         os.path.join(ROOT_DIR, 't2-.xml'),
            >>>         os.path.join(ROOT_DIR, 't3-.xml')]
            >>> Verifyta().easy_verify(model_paths, 
            >>>         verify_options=['-t 1 -o 0', '-t 2 -o 0', '-t 2 -o 1'], 
            >>>         num_threads=3)
            >>> Verifyta().easy_verify(model_paths, trace_path=trace_paths, 
            >>>         verify_options='-t 1 -o 0', num_threads=3)        

        Args:
            model_path (str | List[str]): model paths to be verified.
            trace_path (str | List[str], optional): target trace paths, both `.xtr` and `.xml`(DBM) are supported. 
                Defaults to None, which will create `.xtr` path.
            verify_options (str | List[str], optional): verify options that are proveded by `verifyta`, and you can get details by run `verifyta -h` in your terminal.
                If `verify_options` is provides as a single string, all the models will be verified with the same options. Defaults to '-t 1', returning the shortest trace.
            num_threads (int, optional): use multi-threads if is greater than 1. Defaults to 1.

        Returns:
            List[str]: terminal verify results for each `.xml` model.
        """
        # num_threads will be checked in self.verify()

        # check whether trace_path is None
        if trace_path is None:
            if isinstance(model_path, str):
                trace_path = os.path.splitext(model_path)[0] + '.xtr'
            else:
                trace_path = [os.path.splitext(
                    x)[0]+'.xtr' for x in model_path]

        # check model_path is only one model, set parallel loop
        if isinstance(model_path, str):
            model_path = [model_path]
        if isinstance(trace_path, str):
            trace_path = [trace_path]

        len_model_path = len(model_path)
        len_trace_path = len(trace_path)
        # check model_path and trace_path len is same
        if len_model_path != len_trace_path:
            error_info = f'Length of model_path and trace_path are inconsistent\n' \
                         f'model_path: {len_model_path}, trace_path: {len_trace_path}'
            raise ValueError(error_info)

        if verify_options is None:
            verify_options = "-t 1"
        if isinstance(verify_options, str):
            verify_options = [verify_options for _ in range(len_model_path)]

        options = []
        # 构造options然后让self.verify()处理任务
        for i in range(len_model_path):
            trace_i = trace_path[i]
            verify_options_i = verify_options[i]
            if '-t ' not in verify_options_i:
                verify_options[i] += ' -t 1'
            # model_path existence will be checked in self.verity()
            # check trace_path format
            if trace_i.endswith('.xml'):
                option = f"-X {trace_i.replace('.xml', '')} {verify_options_i}"
                options.append(option)
            elif trace_i.endswith('.xtr'):
                option = f"-f {trace_i.replace('.xtr', '')} {verify_options_i}"
                options.append(option)
            else:
                error_info = f'trace_path should end with ".xml" or ".xtr", ' \
                             f'current trace_path = {trace_i}.'
                raise ValueError(error_info)
        return self.verify(model_path=model_path, verify_options=options, num_threads=num_threads)



# author: Yining She

import re

def get_str_before(string: str, label: str):
    """
    :return: the substring of 'string' before the first occurrence of 'label'.

    If 'label' doesn't exist in 'string', raise error.
    """
    index = string.find(label)
    assert (index != -1)
    return string[:index]


def get_str_after(string, label):
    """
    :return: the substring of 'string' after the first occurence of 'label'.

    If 'label' doesn't exist in 'string', raise error.
    """
    index = string.find(label)
    assert (index != -1)
    return string[index + len(label):]


class Node:
    """
    A instance of a Xml Template.

    :param strname: the name of the instance
    :param str temp_name: the name of the template
    :param list sig_in: a list of names of signals enter the node
    :param list sig_out: a list of names of signals sent by the node
    """

    def __init__(self, name, temp_name, sig_in, sig_out):
        self.name = name
        self.temp_name = temp_name
        self.sig_in = sig_in
        self.sig_out = sig_out


class XmlTemplate:
    """
    Used to define a template in the xml.
    Input is the corresponding code in the xml.
    :param str name: name of the template
    :param list sig_in: a list of signals enter the template
    :param list sig_out: a list of signals sent by the template
    :param dict params: relate input signal variable with its index in the input arguments
    """

    def __init__(self, code):
        self.name = None
        self.signal_in = None
        self.signal_out = None
        self.paras = None
        self.code = code
        self.init_name()
        self.get_all_sync()
        self.get_param()

    def init_name(self):
        """
        Get the name from self.code
        """
        paragraphs = self.code.split('</name>')
        line = paragraphs[0]
        self.name = get_str_after(line, '>')
        # name_start_i = line.index('>')
        # self.name = line[name_start_i+1:]

    def get_all_sync(self):
        """
        Using label "synchronisation" to find all signal names used in the template.
        """
        lines = [i for i in re.findall(r'<label[^<]+</label>', self.code) if "<label kind=\"synchronisation\"" in i]
        self.signal_in = []
        self.signal_out = []
        for line in lines:
            sub = get_str_after(line, '>')
            sub = get_str_before(sub, '</label>')
            if sub[-1] == '?':
                if sub[:-1] not in self.signal_in:
                    self.signal_in.append(sub[:-1])
            else:
                if sub[:-1] not in self.signal_out:
                    self.signal_out.append(sub[:-1])

    def get_param(self):
        """
        Find the signal variables that come in through input.

        And store the name of them with their index in the input arguments, which can be used to find the actual signal variable when give input argument.
        """
        self.paras = {}
        if self.code.find("</parameter>") != -1:
            para_code = get_str_before(self.code, "</parameter>")
            para_code = get_str_after(para_code, "<parameter>")
            para_code = para_code.split(',')
            for i in range(len(para_code)):
                if 'broadcast chan' in para_code[i]:
                    signal_name = get_str_after(para_code[i], 'broadcast chan &amp;')
                    self.paras[signal_name] = i

    def get_instance(self, name, para_string):
        """
        Return and initialize an instance of the template using the input argument.

        :param str name: name of the new instance
        :param str para_string: code of the input argument
        """
        sig_in = []
        sig_out = []
        paras = para_string.split(',')
        for i in self.signal_in:
            if i in self.paras.keys():
                sig_in.append(paras[self.paras[i]].strip())
            else:
                sig_in.append(i)
        for i in self.signal_out:
            if i in self.paras.keys():
                sig_out.append(paras[self.paras[i]].strip())
            else:
                sig_out.append(i)
        return Node(name, self.name, sig_in, sig_out)


class XmlReader:
    """
    Used to read the xml file and take advantage of classes defined above.
    """

    def __init__(self, filename):
        """
        Read the file and init all templates in it.
        """
        self.templates = None
        self.nodes = None
        self.filename = filename
        # print("Open file:", filename)
        f = open(filename,'rt',encoding='utf-8')
        self.code = f.read()

        # print("Initiate all templates:")
        self.init_template()
        # print("Finish Init Templates.")

        # self.get_notes()

    def init_template(self):
        """
        Find all templates in the code and init a 'XmlTemplate' for each of them.
        """
        self.templates = {}
        paragraphs = self.code.split('<template>')
        if len(paragraphs) == 1:
            # print("  No template in file.")
            return
        paragraphs[-1] = get_str_before(paragraphs[-1], '</template>')
        # paragraphs[-1] = paragraphs[-1][:paragraphs[-1].index('</template>')]
        paragraphs = paragraphs[1:]
        for para in paragraphs:
            temp = XmlTemplate(para)
            temp_name = temp.name
            self.templates[temp_name] = temp
            # print("  Add template: ", temp_name)
            # print("    signal in : ", temp.signal_in)
            # print("    signal out: ", temp.signal_out)
            # print("    all paras : ", list(temp.paras.keys()))

    def get_notes(self):
        """
        By reading the 'system' part in the xml to init all time-automata instances.
        :return: a list fo 'Node'.
        """
        self.nodes = {}
        system_code = get_str_after(self.code, '<system>')
        system_code = get_str_before(system_code, '</system>')
        lines = system_code.split('\n')
        for line in lines:
            if line[:7] == 'system ':
                variables = get_str_before(line[7:], ';').split(',')
                for var_name in variables:
                    var_name = var_name.strip()
                    assert (len(var_name) > 0)

                    if var_name in self.templates.keys():
                        self.nodes[var_name] = self.templates[var_name].get_instance(var_name, '')
                    else:
                        l = ''
                        for i in range(len(lines)):
                            if lines[i][:len(var_name)] == var_name:
                                l = get_str_after(lines[i], '=')
                                break
                        assert (l != '')
                        temp_name = get_str_before(l, '(').strip()
                        para_string = get_str_before(l, ')')
                        para_string = get_str_after(para_string, '(')
                        self.nodes[var_name] = self.templates[temp_name].get_instance(var_name, para_string)
                    # print('  Add Node: ', var_name, ' --- template: ', self.nodes[var_name].temp_name)
                    # print('    in:', self.nodes[var_name].sig_in, '  out:', self.nodes[var_name].sig_out)
        return self.nodes


def add_edge(a, b, label=''):
    code = a + '--' + label + '-->' + b
    return code


def build_cg_code(nodes):
    """
    Given a list of 'Node', build the CG graph and print the markdown code.
    """
    codes = ['graph TD']
    node_names = list(nodes.keys())
    codes = codes + node_names

    n = len(node_names)
    for i in range(n):
        for j in range(i + 1, n):
            if i == j:
                continue
            node1 = nodes[node_names[i]]
            node2 = nodes[node_names[j]]
            for signal in node1.sig_out:
                if signal in node2.sig_in:
                    codes.append(add_edge(node1.name, node2.name, signal))
            for signal in node2.sig_out:
                if signal in node1.sig_in:
                    codes.append(add_edge(node2.name, node1.name, signal))

    # Merge all lines of code
    code = '\n'.join(codes)
    return code


def build_cg(filename):
    reader = XmlReader(filename)
    nodes = reader.get_notes()
    code = build_cg_code(nodes)
    mermaid = """```mermaid\n{}```""".format(code)
    return mermaid



"""umodel
"""
# support return typing UModel
from __future__ import annotations
import os
import xml.etree.ElementTree as ET
from typing import List, Tuple
from .datastruct import TimedActions
from .verifyta import Verifyta
from .iTools import UFactory, build_cg, Mermaid
from .tracer import SimTrace, Tracer


class UModel:
    """Load UPPAAL model for analysis, editing, verification and other operations.
    """

    def __init__(self, model_path: str, auto_save=True):
        """_summary_

        Args:
            model_path (str): _description_
            auto_save (bool, optional): whether auto save the model after each operation. Defaults to True.
        """
        self.__model_path: str = model_path
        self.__element_tree: ET.ElementTree = ET.ElementTree(
            file=self.model_path)
        self.__root_elem: ET.Element = self.__element_tree.getroot()
        self.auto_save: bool = auto_save

    @property
    def model_path(self) -> str:
        """Current model path.

        Returns:
            str: Current model path.
        """
        return self.__model_path

    def save_as(self, new_model_path: str) -> UModel:
        """Save the model to a new path with `self.model_path` changed to `new_model_path`.

        Args:
            new_model_path (str): target model path.

        Returns:
            UModel: self.
        """
        with open(new_model_path, 'w', encoding='utf-8') as f:
            self.__element_tree.write(
                new_model_path, encoding="utf-8", xml_declaration=True)
        self.__model_path = new_model_path
        return self

    def save(self) -> UModel:
        """Save the current model.

        Returns:
            UModel: self.
        """
        return self.save_as(self.model_path)

    def copy_as(self, new_model_path: str) -> UModel:
        """Make a copy of the current model and return the copied instance.

        Args:
            new_model_path (str): target copy file path.

        Returns:
            UModel: copied instance.
        """
        with open(new_model_path, 'w', encoding='utf-8') as f:
            self.__element_tree.write(
                new_model_path, encoding="utf-8", xml_declaration=True)
        return UModel(new_model_path)

    def get_communication_graph(self, save_path=None, is_beautify=True) -> Mermaid:
        """Get the communication graph of the UPPAAL model, and return a `Mermaid` instance.

        Args:
            save_path (_type_, optional): `<.md | .svg | .pdf | .png>`, the path to save the file. Defaults to None.
            is_beautify (bool, optional): whether beautify the mermaid file by merging edges. Defaults to True.

        Returns:
            Mermaid: _description_
        """
        mermaid_str = build_cg(self.model_path)
        m = Mermaid(mermaid_str)
        if is_beautify:
            m.beautify()
        if save_path:
            m.export(save_path)
        return m

    def verify(self, trace_path: str = None, verify_options: str = None) -> List[str]:
        """Verify and return the verify result. If `trace_path` is not given, it wll return the list of terminal result.

        Args:
            trace_path (str, optional): _description_. Defaults to None.
            verify_options (str, optional): _description_. Defaults to None.

        Returns:
            List[str]: _description_
        """
        if trace_path:
            return Verifyta().easy_verify(self.model_path, trace_path, verify_options)[0]
        else:
            return Verifyta().verify(self.model_path, verify_options)[0]

    def easy_verify(self, verify_options: str = "-t 1") -> SimTrace | None:
        """Easily verify current model, create a `.xtr` trace file that has the same name as `self.model_path`, and return the SimTrace (if exists).

        Args:
            verify_options (str, optional): verify options, and `-t` must be set because returning a `SimTrace` requires a `.xtr` trace file. Defaults to '-t 1', returning the shortest trace.

        Returns:
            SimTrace | None: if exists a counter example, return a SimTrace, else return None.
        """
        if '-t' not in verify_options:
            err_info = '"-t" must be set in verify_options, '
            err_info += f'current verify_options: {verify_options}.'
            raise ValueError(err_info)
        xtr_trace_path = self.model_path.replace('.xml', '.xtr')

        verify_cmd_res = Verifyta().easy_verify(
            self.model_path, xtr_trace_path, verify_options=verify_options)[0]

        xtr_trace_path = xtr_trace_path.replace('.xtr', '-1.xtr')
        if os.path.exists(xtr_trace_path):
            return Tracer.get_timed_trace(self.model_path, xtr_trace_path)
        else:
            return None

    # ======== templates ========
    @property
    def templates(self) -> List[str]:
        """Get all template names of current model.

        Returns:
            List[str]: list of all template names.
        """
        template_names = self.__element_tree.findall("./template/name")
        return [i.text for i in template_names]

    def remove_template(self, template_name: str) -> bool:
        """
        Delete the template according to the input name.

        :param str template_name: the name of template
        :return: `True` when succeed, `False` when fail
        """
        # get template
        template_elem = None
        for template in self.__element_tree.iter("template"):
            if template.find('name').text == template_name:
                template_elem = template
        # remove template
        if template_elem is None:
            if self.auto_save:
                self.save()
            return False
        self.__root_elem.remove(template_elem)
        if self.auto_save:
            self.save()
        return True

    # ======== queries ========
    @property
    def queries(self) -> List[str]:
        """Get all queries string.

        Returns:
            List[str]: _description_
        """
        query_formula_elems = self.__element_tree.findall(
            './queries/query/formula')
        queries = [query_elem.text for query_elem in query_formula_elems]
        return queries

    def clear_queries(self) -> bool:
        """Clear all queries of the model.

        Returns:
            bool: _description_
        """
        root = self.__root_elem
        queries_elem = root.find('queries')
        if queries_elem is None:
            if self.auto_save:
                self.save()
            return False
        root.remove(queries_elem)
        if self.auto_save:
            self.save()
        return True

    def set_queries(self, queries: List[str] | str) -> None:
        """Delete all the queries in the model and then inserts the new queries into the model

        Args:
            queries (List[str] | str): A list of queries or a single query

        """

        if isinstance(queries, str):
            queries = [queries]

        # 首先删除所有的queries
        self.clear_queries()
        # 然后构造queries并插入到模型中
        queries_elem = UFactory.queries(queries)
        self.__root_elem.append(queries_elem)
        if self.auto_save:
            self.save()

    # ======== system ========
    @property
    def system(self) -> str:
        """Get the system of the model.

        Returns:
            str: _description_
        """
        system_elem = self.__element_tree.find('system')
        return system_elem.text

    def set_system(self, system_str: str) -> None:
        """Set the system element of current model.

        Args:
            system_str (str): target system string you want to set.
        """
        system_elem = self.__element_tree.find('system')
        system_elem.text = system_str
        if self.auto_save:
            self.save()

    # ======== declaration ========
    @property
    def declaration(self) -> str:
        """Get the declaration of the model.

        Returns:
            str: _description_
        """
        declaration_elem = self.__element_tree.find('declaration')
        return declaration_elem.text

    def set_declaration(self, declaration_str: str) -> None:
        """Set the declaration element of current model.

        Args:
            declaration_str (str): target declaration string you want to set.
        """
        declaration_elem = self.__element_tree.find('declaration')
        declaration_elem.text = declaration_str
        if self.auto_save:
            self.save()

    # ======== other ========
    @property
    def __max_location_id(self) -> int:
        """Get the maximum location_id so as to make it easier to create a new template.

        Returns:
            int: max location id.
        """
        location_elems = self.__element_tree.findall('./template/location')
        # <location id="id0" x="-187" y="-76">
        # <location id="id1" x="25" y="-76">
        # <location id="id2" x="-51" y="-119">
        ids = [int(location_elem.attrib['id'][2:])
               for location_elem in location_elems]
        return max(ids)

    @property
    def broadcast_chan(self) -> List[str]:
        """Get broadcast channels in Declaration.

        Returns:
            List[str]: List of broadcast channels.
        """
        declarations = self.declaration
        systems = self.system
        start_index = 0
        broadcast_chan = []
        while True:
            start_index = declarations.find('broadcast chan', start_index, -1)
            if start_index == -1:
                break
            end_index = declarations.find(';', start_index, -1)
            tmp_actions = declarations[start_index +
                                       15:end_index].strip().split(',')
            tmp_actions = [x.strip() for x in tmp_actions]
            broadcast_chan += tmp_actions
            start_index = end_index
        start_index = 0
        return list(set(broadcast_chan))

    def add_observer_template(self, observations: TimedActions, focused_actions: List[str] | None = None, template_name: str = 'Observer', is_strict: bool = True) -> None:
        """Add an observer template, which will also be embedded in `system declarations`. Template that has the same name will be over written.

        An observer is xxx.

        Args:
            observations (TimedActions): observed actions, observed time lower_bound, observed time upper_bound.
            template_name (str, optional): the name of the template. Defaults to 'observer'.
            is_strict (bool, optional): if strict, any other observations will be illegal. 
                For example, assume you set observations `a1, gclk=1, a2, gclk=3`, and there exists trace T: `a1, gclk=1, a2, gclk=2, a2, gclk=3`.
                If `is_strict_observer` is True, then T is invalid. Defaults to True.

        Returns:
            _type_: _description_
        """
        if focused_actions is None:
            focused_actions = list(map(lambda x: x.replace(
                '!', '').replace('?', ''), observations.actions))
        self.add_monitor_template(
            template_name, observations, focused_actions, strict=is_strict)

    def add_pattern_template(self, pattern_list: List[str], template_name: str) -> None:
        """Add a pattern template, which will also be embedded in `system declarations`. Template that has the same name will be over written.

        Args:
            pattern_list (List[str]): pattern to be monitored.
            template_name (str): _description_

        Returns:
            _type_: _description_
        """
        raise NotImplementedError

    def add_input_template(self, signals: TimedActions, template_name: str = 'Input') -> None:
        """Add a linear input template, which will also be embedded in `system declarations`. Template that has the same name will be over written.

        Args:
            signals (TimedActions): _description_
            template_name (str, optional): _description_. Defaults to 'Input'.

        Returns:
            _type_: _description_
        """
        assert len(signals) > 0

        start_id = self.__max_location_id + 1
        # 删除相同名字的monitor
        self.remove_template(template_name)

        clock_name, signals = self.__parse_signals(signals)
        input_model = UFactory.input(
            template_name, signals.convert_to_list_tuple(clock_name), start_id)
        self.__root_elem.insert(-2, input_model)

        # 将新到monitor加入到system中
        self.__add_template_to_system(template_name)

        return None

    def __add_template_to_system(self, template_name: str):
        """Add a template to system declarations.

        Args:
            template_name (str): the name of the template.

        """
        system_lines: List[str] = self.system.split('\n')
        system_lines: List[str] = list(map(lambda x: x.strip(), system_lines))
        for i, line in enumerate(system_lines):
            if line.strip().startswith('system'):
                system_items = list(
                    map(lambda s: s.strip(), line.strip()[6:-1].split(',')))
                if template_name not in system_items:
                    system_items.append(template_name)
                system_lines[i] = f"system {', '.join(system_items)};"
                break

        self.set_system('\n'.join(system_lines))

    # all patterns
    def add_monitor_template(self, monitor_name: str, signals: TimedActions, focused_actions: List[str] = None, strict: bool = True, allpattern: bool = False):
        """Add new linear monitor template, which will also be embedded in `system declarations`. 

        If `monitor_name` already exists in current templates, it will be overwritten.

        Args:
            monitor_name (str): the template name of the monitor.
            signals (TimedActions): actions, lower_bound, upper_bound.
            focused_actions (List[str], optional): _description_. Defaults to None.
            strict (bool, optional): _description_. Defaults to True.
            allpattern (bool, optional): _description_. Defaults to False.

        Returns:
            _type_: _description_
        """

        # 处理focused_actions is None的情况
        if focused_actions is None:
            focused_actions = self.broadcast_chan

        clock_name, signals = self.__parse_signals(signals)

        if '?' in "".join(focused_actions) or '!' in "".join(focused_actions):
            raise ValueError(
                f"focused_actions should not contain '?' or '!', current focused_actions: {focused_actions}")

        start_id = self.__max_location_id + 1
        # 删除相同名字的monitor
        self.remove_template(monitor_name)
        monitor = UFactory.monitor(monitor_name, signals.convert_to_list_tuple(
            clock_name), focused_actions, start_id, strict, allpattern)
        self.__root_elem.insert(-2, monitor)
        # 将新到monitor加入到system中

        self.__add_template_to_system(monitor_name)

    def __parse_signals(self, signals: TimedActions, default_name: str = "input_clk") -> Tuple[str, TimedActions]:
        """Parse the signals, if the signal name is not specified, then use the default name.

        Args:
            signals (TimedActions): the signals to be parsed.
            default_name (str): the default clock name for the signals.

        Returns:
            Tuple[str, TimedActions]: the clock name and the parsed signals.
        """

        parsed_actions = list(map(lambda x: x.replace(
            '?', '').replace('!', ''), signals.actions))

        if len(signals) > 0 and not isinstance(signals.lb[0], str):
            signals.lb = list(map(str, signals.lb))

        if len(signals) > 0 and not isinstance(signals.ub[0], str):
            signals.ub = list(map(str, signals.ub))

        if len(signals) > 0 and signals.lb[0].strip().find('>') > 0 and signals.lb[0].strip()[0] != '>':
            # If the stmt has clk name, like 'a1 > 1', then use the clk name
            # Only accept input like "a1 > 1" or "a1>1", not ">1" or "1"
            clock_name = signals.lb[0].split('>')[0]
        else:  # Otherwise, we use the default clock name
            # accept input like ">1" or "1"
            clock_name = default_name

        len_lb = len(signals.lb)
        for i in range(len_lb):  # Remove the clock name and operator
            # Note that '>=' must be removed first, otherwise '>=' will be removed to '='
            signals.lb[i] = signals.lb[i].replace(
                clock_name, '').replace('>=', '').replace('>', '').strip()
            signals.ub[i] = signals.ub[i].replace(
                clock_name, '').replace('<=', '').replace('<', '').strip()

        return clock_name, TimedActions(parsed_actions, signals.lb, signals.ub)

    def __find_a_pattern(self, focused_action: List[str] = None, hold: bool = True, options: str = None) -> SimTrace | None:
        """Find a pattern in the current model.

        Args:
            focused_action (List[str], optional): the actions that we want to focus on. Defaults to None.
            hold (bool, optional): whether to keep the temp file. Defaults to True.
            options (str, optional): options for the verifier. Defaults to None.

        Returns:
            SimTrace | None: the founded patterns. None if no pattern is found.
        """
        self.save()
        if options is not None:
            sim_trace = self.easy_verify(options)
        else:
            sim_trace = self.easy_verify()

        if sim_trace is None:
            return None

        trace_path = os.path.splitext(self.model_path)[0] + '-1.xtr'
        pattern_seq = sim_trace.filter_by_actions(focused_action)

        if not hold:
            os.remove(trace_path)
            os.remove(self.model_path)

        return pattern_seq

    def find_all_patterns(self, focused_actions: List[str] = None,
                          hold: bool = True,
                          max_patterns: int = None) -> List[SimTrace]:
        """Find all patterns of the first query in the model.

        Args:
            focused_actions (List[str], optional): the actions that we want to focus on. Defaults to None.
            hold (bool, optional): whether to keep the temp files. Defaults to True.
            max_patterns (int, optional): the maximum number of patterns to find. If None, then all patterns will be found. Defaults to None.

        Returns:
            List[SimTrace]: the list of patterns.
        """
        queries = self.queries
        if len(queries) == 0:
            return []
        all_patterns = self.__find_all_patterns_of_a_query(
            queries[0], focused_actions, hold, max_patterns)
        return all_patterns

    def __find_all_patterns_of_a_query(self, query: str = None,
                                       focused_actions: List[str] = None,
                                       hold: bool = True,
                                       max_patterns: int = None) -> List[SimTrace] | None:
        """Find all patterns that satisfy the query

        Args:
            query (str, optional): the query to be verified. Defaults to None.
            focused_actions (List[str], optional): the actions that we want to focus on. Defaults to None.
            hold (bool, optional): whether to keep the temp files. Defaults to True.
            max_patterns (int, optional): the maximum number of patterns to find. If None, then all patterns will be found. Defaults to None.

        Raises:
            NotImplementedError: only support E<> and A[] queries. Raise when other queries are given

        Returns:
            List[SimTrace]: a list of patterns that satisfy the query.
        """
        # 首先
        query = query.strip()
        if not (query.startswith('A[]') or query.startswith('E<>')):
            raise NotImplementedError('Only support E<> and A[] query!')

        if query.startswith('A[]'):
            default_query = query[3:].strip()
            if default_query[0] == '!':
                default_query = "E<> " + default_query[1:].strip()
            elif default_query[0:3] == "not":
                default_query = "E<> " + default_query[3:].strip()
            else:
                default_query = "E<> ! " + default_query.strip()
        else:
            default_query = query

        new_model_path = os.path.splitext(self.model_path)[0] + '_pattern.xml'
        new_umodel = self.copy_as(new_model_path=new_model_path)

        new_umodel.set_queries(default_query)
        new_patterns = new_umodel.__find_a_pattern(
            focused_actions, hold=hold)  # Keep the temp files until the end

        if new_patterns is None:
            return []

        query_str = default_query
        # 根据初始的pattern构建monitor并循环, 初始Moniter为0
        all_patterns = []
        monitor_id = 0
        iter_ = 1
        while new_patterns is not None:
            all_patterns.append(new_patterns)

            if max_patterns is not None and iter_ >= max_patterns:
                break

            monitor_id += 1
            # 将pattern[List] -> TimedActions
            new_observes = TimedActions(new_patterns.actions)
            new_umodel.add_monitor_template(f'Monitor{monitor_id}', new_observes,
                                            focused_actions=focused_actions, strict=True, allpattern=True)

            # 构造验证语句
            # 构造monitor.pass
            # E<> Monitor0.pass & !Monitor1.pass
            query_str = ' && '.join(
                [f'!Monitor{i}.pass' for i in range(1, monitor_id+1)])
            # E<> !Monitor0.pass & !Monitor1.pass
            query_str = f'{default_query} && {query_str}'

            new_umodel.set_queries(query_str)
            new_patterns = new_umodel.__find_a_pattern(
                focused_actions, hold=hold)
            # Keep the temp files until the end

            if new_patterns is None:
                break

            trace_path = os.path.splitext(new_umodel.model_path)[0] + '-1.xtr'

            iter_ = iter_ + 1

        if not hold:
            os.remove(new_model_path)
            os.remove(trace_path)

        return all_patterns



"""_summary_
"""
from typing import List


class TimedActions:
    """ A serie of actions with lower bounds and upper bounds.
    """

    def __init__(self, actions: List[str], lb: List[str] = None, ub: List[str] = None):
        """_summary_

        Args:
            actions (List[str]): a list of actions.
            lb (List[str], optional): a list of lowerbounds. Defaults to None.
            ub (List[str], optional): a list of upperbounds. Defaults to None.

        Examples:
            >>> timed_actions = TimedActions(['sigOut!', 'sigOut!'], 
            >>>     ['gclk>=20', 'x>=30'], ['gclk<=20', 'x<=30'])
        """
        self.__actions: List[str] = actions
        self.lb = lb
        self.ub = ub
        if self.lb is None:
            self.lb = [-1 for i in range(len(self.actions))]
        if self.ub is None:
            self.ub = [-1 for i in range(len(self.actions))]

    @property
    def actions(self) -> List[str]:
        """A serie of actions 

        Examples:
            >>> timed_actions = TimedActions(['sigOut!', 'sigOut!'], 
            >>>     ['gclk>=20', 'x>=30'], ['gclk<=20', 'x<=30'])
            >>> timed_actions.actions
            ['sigOut!', 'sigOut!']

        Returns:
            List[str]: a list of actions.
        """
        return self.__actions

    @property
    def is_patterns(self) -> bool:
        """Whether the timed actions are patterns.

        Examples:
            >>> timed_actions = TimedActions(['sigOut!', 'sigOut!'], 
            >>>     ['gclk>=20', 'x>=30'], ['gclk<=20', 'x<=30'])
            >>> timed_actions.is_patterns
            False

        Returns:
            bool: True if the timed actions are patterns.
        """
        return all(self.lb) == -1 and all(self.ub) == -1

    def convert_to_list_tuple(self, clk_name='monitor_clk'):
        """Convert to a list of tuple, each tuple is a timed action.

        Examples:
            >>> timed_actions = TimedActions(['sigOut!', 'sigOut!'], 
            >>>     ['gclk>=20', 'x>=30'], ['gclk<=20', 'x<=30'])
            >>> timed_actions.convert_to_list_tuple()
            [('sigOut!', 'gclk>=20', 'gclk<=20'), ('sigOut!', 'x>=30', 'x<=30')]

        Returns:
            List[Tuple[str, str, str]]: a list of tuple, each tuple is a timed action.
        """

        self.clk_name = clk_name
        res = []
        for i in range(len(self.actions)):
            res.append(
                (self.actions[i], f'{clk_name}>={self.lb[i]}', f'{clk_name}<={self.ub[i]}'))
        return res

    def convert_to_patterns(self):
        """Convert to patterns.

        Examples:
            >>> timed_actions = TimedActions(['sigOut!', 'sigOut!'],
            >>>     ['gclk>=20', 'x>=30'], ['gclk<=20', 'x<=30'])
            >>> timed_actions.convert_to_patterns()
            ['sigOut!', 'sigOut!']

        Returns:
            List[str]: a list of actions.
        """
        return self.actions

    def __len__(self):
        return len(self.actions)



"""_summary_
"""
import os
from pyuppaal import Tracer, UModel, Verifyta
import verifyta_path

def _run_test_tracer(model_path, trace_path):
    """_summary_
    """
    model_path = verifyta_path.bring_to_root(model_path)
    trace_path = verifyta_path.bring_to_root(trace_path)

    Verifyta().easy_verify(model_path)
    sim_trace = Tracer.get_timed_trace(model_path, trace_path)
    assert os.path.exists(trace_path) is True
    os.remove(trace_path)
    return sim_trace

def test_tracer_basic():
    """_summary_
    """
    Verifyta().set_verifyta_path(verifyta_path.VERIFYTA_PATH)
    sim_trace = _run_test_tracer("pedestrian.xml", "pedestrian-1.xtr")
    # ==== save raw SimTrace ====
    raw_sim_trace_path = verifyta_path.bring_to_root('pedestrian-raw.txt')
    sim_trace.save_raw(raw_sim_trace_path)
    assert os.path.exists(raw_sim_trace_path) is True
    os.remove(raw_sim_trace_path)

    # ==== save SimTrace ====
    sim_trace_path = verifyta_path.bring_to_root('pedestrian.txt')
    sim_trace.save(sim_trace_path)
    assert os.path.exists(sim_trace_path) is True
    os.remove(sim_trace_path)

def test_tracer_trim():
    _run_test_tracer("test1.xml", "test1-1.xtr")



"""mermaid
"""
# support for str | List[str]
from __future__ import annotations
from typing import List,Dict


class Mermaid:
    """class Mermaid
    """
    def __init__(self, mermaid_str: str):
        self.__mermaid_str: str = mermaid_str
        # List of `[source, edge_name, target]`
        self.__mermaid_list: List[List[str]] = None

    def __str__(self) -> str:
        return self.__mermaid_str
    
    def __repr__(self) -> str:
        return self.__str__()

    @property
    def mermaid_str(self) -> str:
        """Raw mermaid string.

        Returns:
            str: raw mermaid string.
        """
        return self.__mermaid_str

    @property
    def mermaid_list(self) -> List[List[str]]:
        """List of `[source, edge_name, target]`

        Returns:
            List[List[str]]: _description_
        """
        if self.__mermaid_list is None:
            self.__mermaid_to_list()
        return self.__mermaid_list

    def __mermaid_to_list(self) -> List[List[str]]:
        """Transform `mermaid` into List of `[source, edge_name, target]`.

        FROM:
        ```mermaid
        graph TD
        TrafficLights--cGreen-->Cars
        TrafficLights--cYellow-->Cars
        LV1Pedestrian2--pCrss-->Cars```

        TO:
        [[TrafficLights, cGreen, Cars],
         [TrafficLights, cYellow, Cars],
         [LV1Pedestrian2, pCrss, Cars]]

        Returns:
            List[List[str]]: List of `[source, edge_name, target]`.
        """

        # 去掉开头和结尾的冗余
        mermaid_list = self.mermaid_str.replace('```', '').split('\n')
        # i TrafficLights--cYellow-->Cars 变成 [TrafficLights, cYellow, Cars]
        res = [i.replace('>', '').split('--') for i in mermaid_list if '--' in i]
        self.__mermaid_list = res
        return res

    def __list_to_mermaid(self) -> str:
        """Convert `self.__mermaid_list` to `self.__mermaid_str`.

        Returns:
            str: _description_
        """

        tmp = ''
        for source, edge, target in self.__mermaid_list:
            tmp += source + '--' + edge + '-->' + target + '\n'
        return f"```mermaid\ngraph TD\n{tmp}```"

    def beautify(self, join_str: str = ',') -> str:
        """Beautify Mermaid by merging edges with the same source and target.

        FROM:
        ```mermaid
        graph TD
        TrafficLights--cGreen-->Cars
        TrafficLights--cYellow-->Cars
        LV1Pedestrian2--pCrss-->Cars
        ```

        TO:
        ```mermaid
        graph TD
        TrafficLights--cGreen,cYellow-->Cars
        LV1Pedestrian2--pCrss-->Cars
        ```

        Args:
            join_str (str, optional): join flag. Defaults to ','.

        Returns:
            str: beautified mermaid string.
        """

        # ==== STEP1 构造合并后的边的字典 ====
        # FROM:
        # [['TrafficLights', 'cGreen', 'Cars'],
        #  ['TrafficLights', 'cYellow', 'Cars'],
        #  ['LV1Pedestrian2', 'pCrss', 'Cars']]
        # TO:
        # {"['TrafficLights', 'Cars']": ['cGreen', 'cYellow'],
        #  "['LV1Pedestrian2', 'Cars']": ['pCrss']}
        edges_dict: Dict[str, List[str]] = {}
        for i in self.mermaid_list:
            key = str([i[0], i[2]])
            value = i[1]
            if key in edges_dict:
                edges_dict[key].append(value)
            else:
                edges_dict[key] = [value]
        # remove duplicate
        for key in edges_dict:
            edges_dict[key] = sorted(list(set(edges_dict[key])))

        # ==== STEP2 从字典构造出 mermaid_str ====
        edges_str = ''
        for key, val in edges_dict.items():
            # str转list获取边的两端
            [source, target] = eval(key)
            edge_name = join_str.join(val)
            edges_str += f"{source}--{edge_name}-->{target}\n"
        res = f'''```mermaid\ngraph TD\n{edges_str}```'''
        self.__mermaid_str = res
        return res

    def remove(self, components: str | List[str]) -> str:
        """Remove mermaid lines that contains components.

        Args:
            excluded_components (str | List[str]): components to remove.

        Returns:
            str: mermaid string.
        """
        if isinstance(components, str):
            components = [components]
        for l in self.mermaid_list:
            if set(l) & set(components):
                self.__mermaid_list.remove(l)
        self.__mermaid_str = self.__list_to_mermaid()
        return self.mermaid_str

    def export(self, target_file: str) -> str:
        """Export mermaid to a `<.md | .svg | .pdf | .png>` file.

        Args:
            target_file (str): `<.md | .svg | .pdf | .png>`.

        Returns:
            str: path to the target file.
        """
        if target_file.endswith('.md'):
            with open(target_file, 'w', encoding='utf-8') as f:
                f.write(self.mermaid_str)
        elif target_file.endswith('.svg'):
            # 实现后别忘了去tests/test_mermaid.py里修改63行的测试 ['mermaid_test.md', 'mermaid_test.svg', 'mermaid_test.png']
            raise NotImplementedError
        elif target_file.endswith('.pdf'):
            # 实现后别忘了去tests/test_mermaid.py里修改63行的测试 ['mermaid_test.md', 'mermaid_test.svg', 'mermaid_test.png']
            raise NotImplementedError
        elif target_file.endswith('.png'):
            # 实现后别忘了去tests/test_mermaid.py里修改63行的测试 ['mermaid_test.md', 'mermaid_test.svg', 'mermaid_test.png']
            raise NotImplementedError
        else:
            raise ValueError(f'Invalid target_file: {target_file}.')
        return target_file



"""_summary_
"""
import os
from verifyta_path import ROOT_DIR
from pyuppaal import Mermaid


def test_basic():
    """_summary_
    """
    s = '''```mermaid
TrafficLights--cGreen-->Cars
TrafficLights--cYellow-->Cars
LV1Pedestrian1--pCrss-->Cars
LV1Pedestrian2--pCrss-->Cars```'''

    m = Mermaid(s)
    assert m.mermaid_str == s
    # ==== remove ====
    res = '''```mermaid
graph TD
TrafficLights--cGreen-->Cars
TrafficLights--cYellow-->Cars
LV1Pedestrian2--pCrss-->Cars
```'''
    assert m.remove('LV1Pedestrian1') == res
    res = [['TrafficLights', 'cGreen', 'Cars'],
           ['TrafficLights', 'cYellow', 'Cars'],
           ['LV1Pedestrian2', 'pCrss', 'Cars']]
    assert m.mermaid_list == res
    # ==== beautify ====
    res = '''```mermaid
graph TD
TrafficLights--cGreen,cYellow-->Cars
LV1Pedestrian2--pCrss-->Cars
```'''
    assert m.beautify() == res
    # ==== remove ====
    res = '''```mermaid
graph TD
TrafficLights--cGreen-->Cars
TrafficLights--cYellow-->Cars
```'''
    assert m.remove('LV1Pedestrian2') == res
    # ==== beautify ====
    res = '''```mermaid
graph TD
TrafficLights--cGreen,cYellow-->Cars
```'''
    assert m.beautify() == res


def test_save():
    """_summary_
    """
    s = '''```mermaid
TrafficLights--cGreen-->Cars
TrafficLights--cYellow-->Cars
LV1Pedestrian1--pCrss-->Cars
LV1Pedestrian2--pCrss-->Cars```'''
    m = Mermaid(s)
    export_files = ['mermaid_test.md', 'mermaid_test.svg', 'mermaid_test.pdf', 'mermaid_test.png']
    for export_file in export_files[0:1]:
        export_file = os.path.join(ROOT_DIR, export_file)
        m.export(export_file)
        assert os.path.exists(export_file) is True
        # delete file
        os.remove(export_file)



import platform
import sys
import os

# FILE_ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
# sys.path.append(FILE_ROOT_DIR.join('utap_39.pyd'))
# sys.path.append(FILE_ROOT_DIR.join('utap_39.lib'))
python_version = platform.python_version()

utap_parser = None

if '3.6' in python_version:
    from .utap_36 import parse
elif '3.7' in python_version:
    from .utap_37 import parse
elif '3.8' in python_version:
    from .utap_38 import parse
elif '3.9' in python_version:
    from .utap_39 import parse
elif '3.10' in python_version:
    from .utap_310 import parse
else:
    raise ValueError(f'Only support python 3.6-3.10, current version: {python_version}.')

utap_parser = parse


# def utap_parser(if_str: str, xtr_file: str) -> str:
#     """接口函数来展示utap_parser的用法。

#     Args:
#         if_str (str): uppaal编译出来的if文件的内容
#         xtr_file (str): `.xtr`路径文件

#     Returns:
#         str: 返回解析好的raw trace
#     """
#     raise NotImplemented


import os
import platform

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

WINDOWS_VERIFYTA_PATH = os.path.join(
    ROOT_DIR, 'uppaal\\Win_Linux-uppaal64-4.1.26\\bin-Windows\\verifyta.exe')
LINUX_VERIFYTA_PATH = os.path.join(
    ROOT_DIR, 'uppaal/Win_Linux-uppaal64-4.1.26/bin-Linux/verifyta')
MAC_VERIFYTA_PATH = os.path.join(
    ROOT_DIR, 'uppaal/macOS-uppaal64-4.1.26/bin-Darwin/verifyta')

path_dir = {
    'Windows': WINDOWS_VERIFYTA_PATH,
    'Linux': LINUX_VERIFYTA_PATH,
    'Darwin': MAC_VERIFYTA_PATH
}

VERIFYTA_PATH = path_dir[platform.system()]


def bring_to_root(file_name: str) -> str:
    """Add root prefix to `file_name`.

    Args:
            file_name (str): _description_

    Returns:
            str: _description_
    """
    return os.path.join(ROOT_DIR, file_name)



import os
# from symbol import simple_stmt
import pyuppaal
from pyuppaal import UModel
from verifyta_path import *
from pyuppaal import TimedActions

pyuppaal.set_verifyta_path(VERIFYTA_PATH)


def test_cg():
    """Test communication graph.
    """
    u = UModel(bring_to_root('pedestrian.xml'))
    # ==== before beautify ====
    tmp_file = bring_to_root('cg_pedestrain.md')
    m = u.get_communication_graph(save_path=tmp_file, is_beautify=False)
    assert os.path.exists(tmp_file) is True
    os.remove(tmp_file)
    print('==== before beautify ====\n', m)

    # ==== after beautify ====
    m.beautify()
    tmp_file = bring_to_root('cg_pedestrain_beauty.md')
    m.export(tmp_file)
    assert os.path.exists(tmp_file) is True
    os.remove(tmp_file)
    print('==== after beautify ====\n', m)

def test_all_patterns():
    """_summary_
    """
    query = f'E<> (PPedestrian.Crossing and PCar.Crossing)' # property query
    focused_actions = ["pCheckLight", "pGreen", "pRed", "pYellow", "pCrss", "cCrss"]
    u = UModel(bring_to_root('pedestrian_new.xml'), auto_save=False)
    u.set_queries([query])
    res = u.find_all_patterns(focused_actions)
    assert len(res) == 4
    # print(len(res), list(map(lambda x: x.actions, res)))



"""A research tool that can simulate, verify or modify UPPAAL models with 
python. It can also help to analyze counter-examples in .xml format
"""

from .verifyta import Verifyta
from .umodel import UModel
from .tracer import ClockZone, Transition, SimTrace, Tracer, GlobalVar
# from .config import *
from .pyuppaal import *
from .iTools import Mermaid
from .utap import utap_parser
from .datastruct import TimedActions

__version__='1.0.3'



import setuptools
import re

def get_property(prop, project):
    result = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format(prop), open('./src/pyuppaal/__init__.py').read())
    return result.group(1)


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pyuppaal",
    version = get_property('__version__', 'pyuppaal'),
    author="Jack0Chan",
    author_email="",
    description="A research tool that can simulate, verify or modify UPPAAL models with python. It can also help to analyze counter-examples in .xml format",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Jack0Chan/pyuppaal",
    project_urls={
        "Bug Tracker": "https://github.com/Jack0Chan/pyuppaal/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
    include_package_data=True,
    install_requires=[
        #   'myst_parser',# 'Sphinx -> markdown'
          # ^^^ Not sure if this is needed on readthedocs.org
          # 'something else?',
          ],
)


"""使用方法 ./trace_custom.exe xxx.if xxx.xtr your_output.txt
注意输入要是LF行尾, 不能是CRLF, 否则会报错unknown section
"""
# 这一行的import能够指定class的method返回自身类
# 参考链接：https://www.nuomiphp.com/eplan/11188.html
from __future__ import annotations
from dataclasses import replace
import os
from typing import Dict, Generator, List
from xml.etree.ElementTree import Element, ElementTree
from .verifyta import Verifyta
import xml.etree.cElementTree as ET
from .utap import utap_parser


class OneClockZone:
    """_summary_
    """

    def __init__(self, clock1: str, clock2: str, is_equal: bool, bound_value: int):
        """clock1 - clock2 < or ≤ bound_value

        Args:
            clock1 (str): _description_
            clock2 (str): _description_
            is_equal (bool): _description_
            bound_value (int): _description_
        """
        self.__clock1: str = clock1
        self.__clock2: str = clock2
        self.__is_equal: bool = is_equal
        self.__bound_value: int = bound_value

    def __str__(self):
        sign = '≤' if self.is_equal else '<'
        res = f'{self.clock1} - {self.clock2} {sign} {self.bound_value}'
        return res

    def __repr__(self):
        return self.__str__()

    @property
    def clock1(self) -> str:
        """clock1 - clock2 < or ≤ bound_value

        Returns:
            str: _description_
        """
        return self.__clock1

    @property
    def clock2(self) -> str:
        """clock1 - clock2 < or ≤ bound_value

        Returns:
            str: _description_
        """
        return self.__clock2

    @property
    def is_equal(self) -> bool:
        """clock1 - clock2 < or ≤ bound_value

        Returns:
            str: _description_
        """
        return self.__is_equal

    @property
    def bound_value(self) -> int:
        """clock1 - clock2 < or ≤ bound_value

        Returns:
            str: _description_
        """
        return self.__bound_value


class ClockZone:
    """Composed by `List` of `OneClockZone`.
    """

    def __init__(self, clockzones: List[OneClockZone]):
        """Composed by `List` of `OneClockZone`.

        For the `.xtr` trace, each state contains several clock zones.

        Args:
            clockzones (List[OneClockZone]): _description_
        """
        self.__clockzones: List[OneClockZone] = clockzones

    def __str__(self):
        if self.clockzones:
            res = '['
            for clock_zone in self.clockzones:
                res += f'{clock_zone.__str__()}; '
            res += ']'
            return res

    def __repr__(self):
        return self.__str__()

    @property
    def clockzones(self) -> List[OneClockZone]:
        """`List` of `OneClockZone`.

        Returns:
            List[OneClockZone]: _description_
        """
        return self.__clockzones


class Edges:
    """`Edges` are components of `Transition`. One Transition contains at least one Edge. 
    If more than 1 Edges are contained, it means that there is sync occurred. On this condition, Edges[0] is the sender(!) and others are reveivers(?).

    An Edge looks like this: 
    process.start_location -> process.end_location: {guard, sync, update}.
    A Transition looks like this:
    Transition: LV1Pedestrian2.Idle -> LV1Pedestrian2.CheckTL {1; pWantCrss!; 1;} TrafficLights.cRed_pGreen -> TrafficLights._id8 {1; pWantCrss?; 1;}
    """

    def __init__(self, start_location: str, end_location: str, guard: str, sync: str, update: str):
        """process.start_location -> process.end_location: {guard, sync, update}

        Args:
            start_location (str): _description_
            end_location (str): _description_
            guard (str): _description_
            sync (str): _description_
            update (str): _description_
        """
        self.__start_location: str = start_location
        self.__end_location: str = end_location
        self.__guard: str = guard
        self.__sync: str = sync
        self.__update: str = update
        # one edge has only one process
        self.__process: str = self.start_location.split('.')[0]

    def __str__(self):
        res = f'{self.process}.{self.start_location} -> {self.process}.{self.end_location}:'
        res = res + f'{{{self.guard};{self.sync};{self.update}}}'
        return res

    def __repr__(self):
        return self.__str__()

    @property
    def start_location(self) -> str:
        """_summary_

        Returns:
            str: _description_
        """
        return self.__start_location

    @property
    def end_location(self) -> str:
        """_summary_

        Returns:
            str: _description_
        """
        return self.__end_location

    @property
    def guard(self) -> str:
        """If `guard == '1'`, it means the edge does not have a guard.

        Returns:
            str: _description_
        """
        return self.__guard

    @property
    def has_guard(self) -> bool:
        """Whether the edge has a gurad.

        If `guard == '1'`, it means the edge does not have a guard.

        Returns:
            bool: _description_
        """
        return self.guard == '1'

    @property
    def sync(self) -> str:
        """_summary_

        Returns:
            str: _description_
        """
        return self.__sync

    @property
    def update(self) -> str:
        """_summary_

        Returns:
            str: _description_
        """
        return self.__update

    @property
    def process(self) -> str:
        """One edge has only one process.

        Returns:
            str: _description_
        """
        return self.__process

    @property
    def is_sync(self) -> bool:
        """_summary_

        Returns:
            bool: _description_
        """
        return self.sync[-1] == '!' or self.sync[-1] == '?'

    @property
    def sync_type(self) -> str:
        """_summary_

        Returns:
            str: _description_
        """
        if self.is_sync:
            return 'send' if self.sync[-1] == '!' else 'receive'
        else:
            return None

    @property
    def sync_symbol(self) -> str:
        """_summary_

        Returns:
            str: _description_
        """
        if self.is_sync:
            return self.sync[:-1]
        else:
            return None


class Transition:
    """One `Transition` contains at least one `Edge`. 
    If more than 1 Edges are contained, it means that there is sync occurred. On this condition, Edges[0] is the sender(!) and others are reveivers(?).

    An Edge looks like this: 
    process.start_location -> process.end_location: {guard, sync, update}.
    A Transition looks like this:
    Transition: LV1Pedestrian2.Idle -> LV1Pedestrian2.CheckTL {1; pWantCrss!; 1;} TrafficLights.cRed_pGreen -> TrafficLights._id8 {1; pWantCrss?; 1;}
    """

    def __init__(self, sync: str, start_process: str, end_process: List[str], edges: List[Edges] = None):
        """_summary_

        Args:
            sync (str): _description_
            start_process (str): _description_
            end_process (List[str]): _description_
            edges (List[Edges], optional): _description_. Defaults to None.
        """
        self.__sync: str = sync
        self.__start_process: str = start_process
        self.__end_process: List[str] = end_process
        # 这个edges暂时没用，但是我们也先存下来了。
        self.__edges: List[Edges] = edges

    def __str__(self):
        if self.sync is None:
            res = f'{self.sync}: {self.edges[0].start_location} -> {self.edges[0].end_location}'
        else:
            res = f'{self.sync}: {self.start_process} -> {self.end_process}; '
            # if self.edges is not None:
            for edge in self.edges:
                res = res + f'{edge.start_location} -> {edge.end_location}; '
        return res

    def __repr__(self):
        return self.__str__()

    @property
    def sync(self) -> str:
        return self.__sync

    @property
    def start_process(self) -> str:
        return self.__start_process

    @property
    def end_process(self) -> List[str]:
        return self.__end_process

    @property
    def edges(self) -> List[Edges]:
        return self.__edges

    @property
    def action(self) -> str:
        return self.sync


class GlobalVar:
    def __init__(self, variables_name: List[str], variables_value: List[float]):
        """GlobalVar: variables_name[i] = variables_value[i]

        Args:
            variables_name (List[str]): _description_
            variables_value (List[float]): _description_
        """
        self.variables_name: List[str] = variables_name
        self.variables_value: List[float] = variables_value

    def __str__(self):
        """_summary_

        Returns:
            _type_: _description_
        """
        if self.variables_name:
            res = '['
            for i in range(len(self.variables_name)):
                res += f'{self.variables_name[i]}={self.variables_value[i]}; '
            res += ']'
            return res

    def __repr__(self):
        return self.__str__()


class SimTrace:
    """_summary_
    """

    def __init__(self, trace_string: str):
        """第i个state和第i个transition有相同的clock constraints

        第i个transition前面跟第i个state, 详情看__str__()

        Args:
            trace_string (str): _description_
            parse_raw (bool, optional): whether parse the raw trace string to components. Defaults to True.
        """
        self.__has_parse_raw = False
        self.__raw: str = trace_string
        self.__states: List[List[str]] = None
        self.__global_variables: List[GlobalVar] = None
        self.__clock_constraints: List[ClockZone] = None
        self.__transitions: List[Transition] = None

        # self.__parse_raw()

    def __str__(self):
        """Convert the trace to string.

        Examples:
            >>> trace = SimTrace(trace_from_pedestrian)
            >>> str(trace)
            State [0]: ['Cars.Idle', 'TrafficLights.cRed_pGreen', 'LV1Pedestrian2.Idle']
            global_variables [0]: [Cars.tCCrssMax=4; Cars.tCCrssMin=1; LV1Pedestrian2.tPCrssMin=0; LV1Pedestrian2.tPCrssMax=10; ]
            Clock_constraints [0]: [t(0) - tTL ≤ 0; t(0) - Cars.tc ≤ 0; t(0) - TrafficLights.tTL ≤ 0; t(0) - LV1Pedestrian2.tp ≤ 0; t(0) - LV1Pedestrian2.tTL ≤ 0; tTL - t(0) ≤ 55; tTL - Cars.tc ≤ 0; Cars.tc - TrafficLights.tTL ≤ 0; TrafficLights.tTL - LV1Pedestrian2.tp ≤ 0; LV1Pedestrian2.tp - LV1Pedestrian2.tTL ≤ 0; LV1Pedestrian2.tTL - tTL ≤ 0; ]
            transitions [0]: pWantCrss: LV1Pedestrian2 -> ['TrafficLights']
            -----------------------------------
            State [1]: ['Cars.Idle', 'TrafficLights._id8', 'LV1Pedestrian2.CheckTL']
            global_variables [1]: [Cars.tCCrssMax=4; Cars.tCCrssMin=1; LV1Pedestrian2.tPCrssMin=0; LV1Pedestrian2.tPCrssMax=10; ]
            Clock_constraints [1]: [t(0) - tTL ≤ 0; t(0) - Cars.tc ≤ 0; t(0) - TrafficLights.tTL ≤ 0; t(0) - LV1Pedestrian2.tp ≤ 0; t(0) - LV1Pedestrian2.tTL ≤ 0; tTL - t(0) ≤ 55; tTL - Cars.tc ≤ 0; Cars.tc - TrafficLights.tTL ≤ 0; TrafficLights.tTL - LV1Pedestrian2.tp ≤ 0; LV1Pedestrian2.tp - LV1Pedestrian2.tTL ≤ 0; LV1Pedestrian2.tTL - tTL ≤ 0; ]
            transitions [1]: pGreen: TrafficLights -> ['LV1Pedestrian2']
            -----------------------------------
            State [2]: ['Cars.Idle', 'TrafficLights.cRed_pGreen', 'LV1Pedestrian2._id27']
            global_variables [2]: [Cars.tCCrssMax=4; Cars.tCCrssMin=1; LV1Pedestrian2.tPCrssMin=0; LV1Pedestrian2.tPCrssMax=10; ]
            Clock_constraints [2]: [t(0) - tTL ≤ 0; t(0) - Cars.tc ≤ 0; t(0) - TrafficLights.tTL ≤ 0; t(0) - LV1Pedestrian2.tp ≤ 0; t(0) - LV1Pedestrian2.tTL ≤ 0; tTL - t(0) ≤ 55; tTL - Cars.tc ≤ 0; Cars.tc - TrafficLights.tTL ≤ 0; TrafficLights.tTL - LV1Pedestrian2.tp ≤ 0; LV1Pedestrian2.tp - LV1Pedestrian2.tTL ≤ 0; LV1Pedestrian2.tTL - tTL ≤ 0; ]

        Returns:
            _type_: _description_
        """
        self.__parse_raw()

        res = ''
        for i in range(len(self.__states)):
            res += f'State [{i}]: {self.__states[i].__str__()}\n'
            res += f'global_variables [{i}]: {self.__global_variables[i].__str__()}\n'
            res += f'Clock_constraints [{i}]: {self.__clock_constraints[i].__str__()}\n'
            if i < len(self.__states) - 1:
                res += f'transitions [{i}]: {self.__transitions[i].__str__()}\n'
                res += f'-----------------------------------\n'
        return res

    def __repr__(self):
        return f'SimTrace(...)'

    def __len__(self):
        self.__parse_raw()
        return len(self.actions)

    def __getitem__(self, index: int | slice | List[int]) -> SimTrace:
        """Get the corresponding `SimTrace` by index.

        Args:
            index (int | slice): The indexs for slicing.

        Returns:
            SimTrace: The corresponding sliced SimTrace.
        """
        self.__parse_raw()
        if isinstance(index, (int, slice)):
            new_trace: SimTrace = SimTrace("")
            new_trace.__has_parse_raw = True
            new_trace.__states = self.__states[index]
            new_trace.__clock_constraints = self.__clock_constraints[index]
            new_trace.__transitions = self.__transitions[index]
            new_trace.__global_variables = self.__global_variables[index]
            return new_trace
        else:
            new_trace: SimTrace = SimTrace("")
            new_trace.__has_parse_raw = True
            new_trace.__states = [self.__states[i] for i in index]
            new_trace.__clock_constraints = [
                self.__clock_constraints[i] for i in index]
            new_trace.__transitions = [self.__transitions[i] for i in index]
            new_trace.__global_variables = [
                self.__global_variables[i] for i in index]
            return new_trace

    @property
    def raw(self) -> str:
        """Original raw string of the trace.

        Examples:
            >>> trace = SimTrace(trace_from_pedestrian)
            >>> trace.raw
            State: Cars.Idle TrafficLights.cRed_pGreen LV1Pedestrian2.Idle Cars.tCCrssMax=4 Cars.tCCrssMin=1 LV1Pedestrian2.tPCrssMin=0 LV1Pedestrian2.tPCrssMax=10 t(0)-tTL<=0 t(0)-Cars.tc<=0 t(0)-TrafficLights.tTL<=0 t(0)-LV1Pedestrian2.tp<=0 t(0)-LV1Pedestrian2.tTL<=0 tTL-t(0)<=55 tTL-Cars.tc<=0 Cars.tc-TrafficLights.tTL<=0 TrafficLights.tTL-LV1Pedestrian2.tp<=0 LV1Pedestrian2.tp-LV1Pedestrian2.tTL<=0 LV1Pedestrian2.tTL-tTL<=0
            Transition: LV1Pedestrian2.Idle -> LV1Pedestrian2.CheckTL {1; pWantCrss!; 1;} TrafficLights.cRed_pGreen -> TrafficLights._id8 {1; pWantCrss?; 1;}
            State: Cars.Idle TrafficLights._id8 LV1Pedestrian2.CheckTL Cars.tCCrssMax=4 Cars.tCCrssMin=1 LV1Pedestrian2.tPCrssMin=0 LV1Pedestrian2.tPCrssMax=10 t(0)-tTL<=0 t(0)-Cars.tc<=0 t(0)-TrafficLights.tTL<=0 t(0)-LV1Pedestrian2.tp<=0 t(0)-LV1Pedestrian2.tTL<=0 tTL-t(0)<=55 tTL-Cars.tc<=0 Cars.tc-TrafficLights.tTL<=0 TrafficLights.tTL-LV1Pedestrian2.tp<=0 LV1Pedestrian2.tp-LV1Pedestrian2.tTL<=0 LV1Pedestrian2.tTL-tTL<=0
            Transition: TrafficLights._id8 -> TrafficLights.cRed_pGreen {1; pGreen!; 1;} LV1Pedestrian2.CheckTL -> LV1Pedestrian2._id27 {1; pGreen?; 1;}
            State: Cars.Idle TrafficLights.cRed_pGreen LV1Pedestrian2._id27 Cars.tCCrssMax=4 Cars.tCCrssMin=1 LV1Pedestrian2.tPCrssMin=0 LV1Pedestrian2.tPCrssMax=10 t(0)-tTL<=0 t(0)-Cars.tc<=0 t(0)-TrafficLights.tTL<=0 t(0)-LV1Pedestrian2.tp<=0 t(0)-LV1Pedestrian2.tTL<=0 tTL-t(0)<=55 tTL-Cars.tc<=0 Cars.tc-TrafficLights.tTL<=0 TrafficLights.tTL-LV1Pedestrian2.tp<=0 LV1Pedestrian2.tp-LV1Pedestrian2.tTL<=0 LV1Pedestrian2.tTL-tTL<=0
            Transition: LV1Pedestrian2._id27 -> LV1Pedestrian2.Crossing {1; pCrss!; tp = 0;} Cars.Idle -> Cars.Idle {1; pCrss?; 1;}

        Returns:
            str: Original raw string of the trace.
        """
        return self.__raw

    @property
    def states(self) -> List[List[str]]:
        self.__parse_raw()
        return self.__states

    @property
    def clock_constraints(self) -> List[ClockZone]:
        self.__parse_raw()
        return self.__clock_constraints

    @property
    def transitions(self) -> List[Transition]:
        self.__parse_raw()
        return self.__transitions

    # 这个就是get_untimed_pattern
    @property
    def actions(self) -> List[str]:
        self.__parse_raw()
        return [x.action for x in self.transitions if x.action is not None]

    @property
    def untime_pattern(self) -> List[str]:
        self.__parse_raw()
        return self.actions

    @property
    def global_variables(self) -> List[GlobalVar]:
        self.__parse_raw()
        return self.__global_variables

    def filter_by_actions(self, focused_actions: List[str]) -> SimTrace:
        """Filter the transitions by actions.

        Args:
            focused_actions (List[str], optional): actions that you take cares of.

        Returns:
            SimTrace: _description_
        """
        self.__parse_raw()
        if focused_actions is None:
            return self
        index_array = [i for i in range(
            len(self.transitions)) if self.transitions[i].action in focused_actions]
        return self[index_array]

    def __parse_raw(self) -> None:
        """Parse raw string to components.
        """
        if self.__has_parse_raw:
            return
        trace_text = self.__raw
        trace_text = trace_text.split('\n')
        clock_constraints, states, global_variables, transitions = [], [], [], []
        for tr_ind in range(len(trace_text)):
            state_text, globalvar_name, globalvar_val, clockzones_text, transitions_text = [], [], [], [], []
            if trace_text[tr_ind].startswith('State'):
                tmp_trace_i = trace_text[tr_ind][7:].strip().split(' ')
                for tr_sub in range(len(tmp_trace_i)):
                    if ("=" in tmp_trace_i[tr_sub]) and ('<' not in tmp_trace_i[tr_sub]):
                        # globalvariables
                        var_name, var_val = tmp_trace_i[tr_sub].split('=')
                        globalvar_name.append(var_name.strip())
                        globalvar_val.append(var_val.strip())
                    elif ('<' in tmp_trace_i[tr_sub]) and ('-' in tmp_trace_i[tr_sub]):
                        # clockzones
                        clocks, clk_bound = [
                            x.strip() for x in tmp_trace_i[tr_sub].split('<')]
                        clock1, clock2 = [x.strip() for x in clocks.split('-')]
                        if clk_bound[0] == '=':
                            clk_bound = clk_bound[1:]
                            is_equal = True
                        else:
                            is_equal = False
                        clkzone = OneClockZone(clock1=clock1.strip(), clock2=clock2.strip(' '),
                                               is_equal=is_equal, bound_value=clk_bound.strip(' '))
                        clockzones_text.append(clkzone)
                    else:
                        # state
                        state_text.append(tmp_trace_i[tr_sub].strip())
                clock_constraints.append(ClockZone(clockzones_text))
                states.append(state_text)
                global_variables.append(
                    GlobalVar(globalvar_name, globalvar_val))
            elif trace_text[tr_ind].startswith('Transition'):
                # Transition: PHisAAVFast.Retro -> PHisAAVFast._id15 {t >= tCondMin; actNode1!; t = 0;} NHisA.Rest -> NHisA._id7 {1; actNode?; 1;}
                # [PHisAAVFast.Retro -> PHisAAVFast._id15 {t >= tCondMin; actNode1!; t = 0;
                # NHisA.Rest -> NHisA._id7 {1; actNode?; 1;
                # '']
                tmp_trace_i = trace_text[tr_ind][12:].strip().split('}')[:-1]
                edges_list = []
                end_process = []
                sync_symbol = None
                start_process = None
                for tr_sub in range(len(tmp_trace_i)):
                    # [NHisA.Rest -> NHisA._id7,
                    #  1; actNode?; 1;
                    trans_comp, edge_trans = tmp_trace_i[tr_sub].split('{')
                    start_location, end_location = [
                        x.strip() for x in trans_comp.split('->')]
                    guard, sync, update = [x.strip()
                                           for x in edge_trans.split(';')[:-1]]
                    # strip
                    tmp_edge = Edges(
                        start_location, end_location, guard, sync, update)
                    if tmp_edge.sync_type == 'send':
                        start_process = tmp_edge.process
                        sync_symbol = tmp_edge.sync_symbol
                    elif tmp_edge.sync_type == 'receive':
                        end_process.append(tmp_edge.process)
                    else:
                        start_process = tmp_edge.process
                        end_process.append(tmp_edge.process)
                    edges_list.append(tmp_edge)
                transitions.append(Transition(sync=sync_symbol, start_process=start_process,
                                              end_process=end_process, edges=edges_list))
                # print(transitions[-1].edges)
                # assert transitions[-1].edges != []
                # assert transitions[-1].edges is not None
            else:
                pass
        self.__states = states
        self.__clock_constraints = clock_constraints
        self.__transitions = transitions
        self.__global_variables = global_variables
        self.__has_parse_raw = True

    def save_raw(self, file_name: str) -> None:
        """Save raw data to file.

        Args:
            file_name (str): _description_
        """
        with open(file_name, 'w', encoding='utf-8') as f:
            f.write(self.raw)

    def save(self, file_name: str) -> None:
        """Save self.__str__() to file.

        Args:
            file_name (str): _description_
        """
        with open(file_name, 'w', encoding='utf-8') as f:
            # 这里有parse_raw
            f.write(self.__str__())

    def trim_transitions(self, model_path: str) -> None:
        """It reads the `.xml` file and gets the mapping from formal parameters to real parameters, then it
        replaces the formal parameters in the transitions with the real parameters.

        Args:
            model_path (str): The path of the `.xml` file.

        Raises:
            ValueError: When the `.xml` file is invalid.
        """
        self.__parse_raw()
        element_tree_root: Element | None = ET.ElementTree(
            file=model_path).getroot()
        if element_tree_root is None:
            # When the `.xml` file is invalid.
            raise ValueError(f"Invalid UPPAAL template file")

        # Get template parameters from .xml file
        templates: Generator[Element, None,
                             None] = element_tree_root.iterfind("template")
        param_map: Dict[str, List[str]] = dict()
        for template in templates:
            name_element: Element | None = template.find("name")
            if name_element is None:
                # When the `.xml` file is invalid.
                raise ValueError(f"Invalid UPPAAL template file")

            name: str = name_element.text  # uppaal guaranteed this to have only one
            param_elememt: Element | None = template.find(
                "parameter")  # uppaal guaranteed this to have only one
            if param_elememt is not None:
                param_map[name] = list(map(lambda item: item.strip().split(' ')[-1].replace('&', ''),
                                           param_elememt.text.split(',')))  # remove type and ref annotation
            else:
                param_map[name] = []

        # Get system components defination
        system_element: Element | None = element_tree_root.find("system")
        if system_element is None:
            # When the `.xml` file is invalid.
            raise ValueError(f"Invalid UPPAAL template file")

        system_items: List[str] = list(filter(lambda line: len(line) > 0 and line.find("//") != 0 and line.count(' ') != len(line), # remove comment and empty line
                                    system_element.text.replace("\r\n", "\n").replace('\t', '') .split("\n")))[:-1] # unify "\n", remove "\t" and remove "system"

        source_map: Dict[str, Dict[str, str]] = dict()
        for item in system_items:
            name, constructor = item.replace(";", "").replace(
                " ", "").split('=')  # remove ';' and ' '
            left_brace_index: int = constructor.find('(')
            right_brace_index: int = constructor.find(')')
            # get the name of the constructor(template)
            constructor_name: str = constructor[:left_brace_index]
            real_param_list: List[str] = list(filter(
                lambda param: param != '', constructor[left_brace_index + 1: right_brace_index].split(',')))  # get corresponding param list
            # get constructor param list
            form_param_list: List[str] = param_map[constructor_name]

            # ? I am not sure if default value is allowed in UPPAAL
            # When the `.xml` file is invalid.
            assert len(real_param_list) == len(
                form_param_list), f"Invalid UPPAAL template file"

            if len(real_param_list) != 0:
                item_map: Dict[str, str] = dict()
                for i, real_param in enumerate(real_param_list):
                    # map form param to real param
                    item_map[form_param_list[i]] = real_param
                source_map[name] = item_map
        # print(source_map)

        for i, transition in enumerate(self.__transitions):
            if transition.sync is not None \
                    and transition.start_process in source_map.keys() \
                    and transition.sync in source_map[transition.start_process]:
                # If not in the keys, then it does not have a parameter, so we just skip it
                self.__transitions[i] = Transition(source_map[transition.start_process][transition.sync],
                                                   transition.start_process, transition.end_process, transition.edges)


class Tracer:
    """
    Analyze the `xtr` file generated by verification from command line
    """
    @staticmethod
    def get_timed_trace(model_path: str, xtr_trace_path: str) -> SimTrace | None:
        """Analyze the `xtr_trace_path` trace file generated by `model_path` model and return the instance `SimTrace`.

        The internal process is as following: 
        1. Convert the `model_path` model into a `.if` file.
        2. Analyze `.if` file and the `xtr_trace_path` to get the instance `SimTrace`.
        3. [reference](https://github.com/UPPAALModelChecker/utap).

        Args:
            model_path (str): the path of the `.xml` model file
            xtr_trace_path (str): the path of the `.xtr` trace file

        Returns:
            SimTrace | None: if you want to save the parsed raw trace, you can use SimTrace.save_raw(file_name)
        """
        if_str = Verifyta().compile_to_if(model_path=model_path)
        trace_text = utap_parser(if_str, xtr_trace_path)
        res = SimTrace(trace_text)
        res.trim_transitions(model_path)
        return res



import os
import platform

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

WINDOWS_VERIFYTA_PATH = os.path.join(ROOT_DIR, 'uppaal/Win_Linux-uppaal64-4.1.26/bin-Windows/verifyta.exe')
LINUX_VERIFYTA_PATH = os.path.join(ROOT_DIR, 'uppaal/Win_Linux-uppaal64-4.1.26/bin-Linux/verifyta')
MAC_VERIFYTA_PATH = os.path.join(ROOT_DIR, 'uppaal/macOS-uppaal64-4.1.26/bin-Darwin/verifyta')

path_dir = {
	'Windows': WINDOWS_VERIFYTA_PATH,
	'Linux': LINUX_VERIFYTA_PATH,
	'Darwin': MAC_VERIFYTA_PATH
}

VERIFYTA_PATH = path_dir[platform.system()]
